import os
import sys

from mpi4py import MPI

COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

import numpy as np
import utils as ut
import matBuild as mb
import geometry as geo
import gmshMesh as msh
import petscUtils as lag

if RANK == 0:
    if os.path.exists('results'):
        for file in os.listdir('results'):
            os.remove('results/' + file)
        os.removedirs('results')
        os.mkdir('results')
    else:
        os.mkdir('results')
    files = ['nrmH', 'nrmHSup', 'nrmHDiam', 'nrmHSplit', 'nrmHSmart',
             'pointDiam', 'pointSplit', 'pointSmart', 'nrmHSupra', 'nrmJSupra']
    for f in files:
        if os.path.isfile(f):
            os.remove(f)

verbose  = False
viewPart = False
saveMesh = False
saveNrm  = False

################################################################################
#                                           2D                                 #
################################################################################
if RANK == 0:
    print('---------- 3D ----------')

meshDim = 3
meshFile = 'TODO'

msh.gmshToHDF5(SIZE, RANK, COMM, meshFile, meshDim)
cells, nodes, cellsGlobId, nodesGlobId,\
            nbCellsLoc, nbNodesLoc, edgesDom,\
            physCells, physEdgesDom = msh.readPartMesh('hdf5Mesh', RANK)
if verbose:
    print("Cells local (%d) size %d and total size %d."
              %(RANK, nbCellsLoc, len(cells)))
    sys.stdout.flush()

# Build basic mesh info
geo.setLocalNodeId(cells, nodesGlobId, edgesDom)
edgesNodesId = geo.getNodesEdges(edgesDom, meshDim)
nodesCellsId = geo.cellsAroundNodes(cells, len(nodes))
faces, nbInnerFaces, nbSharedFaces = geo.facesInfo(cells, nodes,
            edgesDom, physEdgesDom, nbCellsLoc, cellsGlobId, meshDim)
if verbose:
    print("(%d) Number of local faces %d and shared %d."
              %(RANK, nbInnerFaces, nbSharedFaces))
    sys.stdout.flush()

# Build mesh info num methods
cellsCenter = geo.getCellsCenter(cells, nodes)
facesNormal = geo.getFacesNormal(faces, nodes, cellsCenter, meshDim)
cellsVol    = geo.getCellsVol(cells, nodes, meshDim)
leastSquareCoeff = geo.lastSquaresWeight(nodes, nodesCellsId, nbNodesLoc,
            cellsCenter, nbCellsLoc, faces, meshDim)

# Diam info and splitted diam info
diamVol, diamNormal = geo.getInnerDiamond(faces, nodes, cellsCenter, meshDim)
splitDVol, splitDNormal = geo.getSplitDiamond(meshDim, faces, nodes,
            cellsCenter, nbInnerFaces+nbSharedFaces)

# Cells, Nodes and Faces numbering
allCellsSize, cellsMap1 = geo.globToLocNumbering(SIZE, RANK, COMM,
            cellsGlobId, nbCellsLoc)
allNodesSize, nodesMap1 = geo.globToLocNumbering(SIZE, RANK, COMM,
            nodesGlobId, nbNodesLoc)

cellsMap3 = ut.scaleMap(cellsMap1, 3)
nodesMap3 = ut.scaleMap(nodesMap1, 3)

### Get min h value
if verbose:
    hmi, hma = geo.minMaxEdgesLength(COMM, cells, nodes, nbCellsLoc)
    if RANK == 0:
        print("-----------------------------------------")
        print("Min h value: %2.8f" %(hmi))
        print("Max h value: %2.8f" %(hma))
        print("-----------------------------------------")
        sys.stdout.flush()
### View the mesh partition
if viewPart:
    debug = np.zeros(nbCellsLoc, np.float64)
    for i in range(nbCellsLoc):
        debug[i] = RANK
    msh.saveMeshEData('results/part', debug, cellsGlobId, 1, 0, RANK)

def addNeumLsq(rule, neuVal):
    nodesStrongPhys = []

    assert(len(edgesDom) == len(physEdgesDom))
    tmp = []
    for iEdg in range(len(edgesDom)):
        edg = edgesDom[iEdg]
        if edg[0] != -1:
            tmp.append((edg[0], physEdgesDom[iEdg]))
        if edg[1] != -1:
            tmp.append((edg[1], physEdgesDom[iEdg]))
    tmp.sort()

    sz = len(tmp)
    if len(tmp) == 0:
        return
    elif len(tmp) > 2:
        for i in range(sz-1):
            if tmp[i][0] != tmp[i+1][0]:
                nodesStrongPhys.append(tmp[i])
            elif tmp[i][1] != tmp[i+1][1]:
                if physRules[tmp[i][1]] > physRules[tmp[i+1][1]]:
                    tmp[i+1] = (tmp[i][0], tmp[i][1])
        if tmp[-2][0] != tmp[-1][0]:
            nodesStrongPhys.append(tmp[-2])
        elif tmp[-2][1] != tmp[-1][1]:
            if physRules[tmp[-2][1]] > physRules[tmp[-1][1]]:
                tmp[-1] = (tmp[-1][0], tmp[-2][-1])
        nodesStrongPhys.append(tmp[-1])

    nbEdgesNode = len(nodesStrongPhys)
    for i in range(nbEdgesNode):
        if nodesStrongPhys[i][1] in neuVal:
            iNd = nodesStrongPhys[i][0]
            nbC = len(nodesCellsId[iNd])
            if len(leastSquareCoeff[iNd]) != 0:
                continue
            for j in range(nbC):
                leastSquareCoeff[iNd].append(1 / nbC)
if False:
    if RANK == 0:
        print('---------------------------------------------')
        print('-           Bi-Domaine Diamond              -')
        print('---------------------------------------------')

    physRules = {
            100001: 0,
            100002: 1,
            100003: 1,
            }
    neuPhys = [100001]
    dirPhys = [100002]
    neuPhys = []
    dirPhys = [100001, 100002]
    addNeumLsq(physRules, neuPhys)

    cellsSup = [1000001]
    cellsAir = [1000002]
    physBd   = [100001, 100002]
    allCells = [(1000001, 1000001), (1000001, 1000002),
                (1000002, 1000001), (1000002, 1000002)]
    fBorderD = geo.rulesSeclectFaces(faces, dirPhys, allCells,
                                     physCells, nbInnerFaces,
                                     nbSharedFaces,  meshDim)
    fBorderN = geo.rulesSeclectFaces(faces, neuPhys, allCells,
                                     physCells, nbInnerFaces,
                                     nbSharedFaces,  meshDim)
    fNotBorder = geo.rulesSeclectFaces(faces, [-1], allCells,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)

    allCells = np.ones(nbCellsLoc, np.int32)
    for i in range(nbCellsLoc):
        allCells[i] = i

    # Space for coeff
    szBuff = 100000000
    row   = np.zeros(szBuff,   np.int32)
    col   = np.zeros(szBuff,   np.int32)
    coeff = np.zeros(szBuff, np.float64)

    t = 0.
    it = 0
    dt = 1e-3
    tEnd = 150*dt

    rhoSup =   1e-3
    rhoAir =   1e-2
    rhoRules = {
            cellsSup[0]: rhoSup,
            cellsAir[0]: rhoAir,
            }

    if False:
        def _fun0(p):
            s = np.zeros(1)
            s[0] = 1.
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            r = rho[rowId]
            vSol = _fun0(pts)
            v = 3 * vD

            row[size] = rowId
            coeff[size] = - r * alpha \
                    * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2]) \
                    * vSol[0] / v
            size += 1

            return 1

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        args):
            fun = args[0]
            rho = args[1][L]

            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   vR * np.dot(nF, nF)
            v += - vA * np.dot(nF, nDA)
            v += - vB * np.dot(nF, nDB)
            v += - vC * np.dot(nF, nDC)
            v  = - v * rho / (3 * vD)

            row[size]    =    L
            coeff[size]  = v[0]
            size        +=    1

            return 1

        buildFun =       mb.lapl3
        idFun    =    mb.cellsId1
        solFun   =          _fun0
        funDiri  =         _diri0
        diriRhs  =    funDiriRhs3
        pbDim    =              1
        pbMap    =      cellsMap1

        kspMaxIt =  500
        ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs',  'asm', 1e-8, 1e-4, kspMaxIt)
    elif False:
        def _fun0(p):
            s = np.zeros(3)
            s[1] = np.cos(p[0])
            s[1] = 1.
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            r = rho[rowId]
            vSol = _fun0(pts)
            v = 3 * vD

            row[size]    =  3*rowId
            coeff[size]  = - alpha * r * (nA[1] * nB[1] + nA[2] * nB[2]) * vSol[0] / v
            coeff[size] +=   alpha * r * (nA[1] * nB[0]) * vSol[1] / v
            coeff[size] +=   alpha * r * (nA[2] * nB[0]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +1
            coeff[size]  = - alpha * r * (nA[0] * nB[0] + nA[2] * nB[2]) * vSol[1] / v
            coeff[size] +=   alpha * r * (nA[0] * nB[1]) * vSol[0] / v
            coeff[size] +=   alpha * r * (nA[2] * nB[1]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +2
            coeff[size]  = - alpha * r * (nA[0] * nB[0] + nA[1] * nB[1]) * vSol[2] / v
            coeff[size] +=   alpha * r * (nA[0] * nB[2]) * vSol[0] / v
            coeff[size] +=   alpha * r * (nA[1] * nB[2]) * vSol[1] / v
            size += 1

            return 3

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        args):
            fun = args[0]
            rho = args[1][L]

            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   np.cross(nF, np.cross( nF, vR))
            v += - np.cross(nF, np.cross(nDA, vA))
            v += - np.cross(nF, np.cross(nDB, vB))
            v += - np.cross(nF, np.cross(nDC, vC))
            v  = v * rho / (3 * vD)

            row[size]    =    3*L
            coeff[size]  =   v[0]
            size        +=      1

            row[size]    = 3*L +1
            coeff[size]  =   v[1]
            size        +=      1

            row[size]    = 3*L +2
            coeff[size]  =   v[2]
            size        +=      1

            return 3

        buildFun =   mb.curlCurl3
        idFun    =    mb.cellsId3
        solFun   =          _fun0
        funDiri  =         _diri0
        diriRhs  =    funDiriRhs3
        pbDim    =              3
        pbMap    =      cellsMap3

        kspMaxIt =  500
        ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs',  'asm', 1e-8, 1e-4, kspMaxIt)
    else:
        raise ValueError("No config selected")

    # Cpt Rho
    def computeRho(rho, nbCellsLoc, JLoc, physCells, rhoDict):
        sz = rho.getSizes()[0]
        arrRho = np.zeros(sz, np.float64)
        for i in range(nbCellsLoc):
            v = rhoDict[physCells[i]]
            arrRho[i] = v
        rho.setArray(arrRho)
        return rho

    ### Ghost vec and init
    ghostId = ut.getForeignId(cellsMap1, nbCellsLoc, 1)
    rho = lag.createGhostVec(nbCellsLoc, 1, 1, ghostId)
    computeRho(rho, nbCellsLoc, rho, physCells, rhoRules)
    lag.syncGhostVec(rho)
    arrayRho = lag.getGhostArray(rho)

    coeffBuild = np.zeros(len(faces), np.float64)
    for i in range(len(faces)):
        f = faces[i]
        L = f[3]
        R = f[4]
        if R != -1:
            r = 0.5 * (arrayRho[L] + arrayRho[R])
        else:
            r = arrayRho[L]
        coeffBuild[i] = r / (3 * diamVol[i])

    # Vec
    H     = lag.createVec(nbCellsLoc, pbDim, pbDim)
    prevH = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsBd = lag.createVec(nbCellsLoc, pbDim, pbDim)
    zero  = lag.createVec(nbCellsLoc, pbDim, pbDim)
    H.assemblyBegin()
    H.assemblyEnd()

    # Mat
    matOP  = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)
    matVol = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)

    ### Build the matrix
    size = mb.ExDiamond(meshDim, row, col, coeff, fNotBorder, faces,
                        facesNormal, diamNormal, diamVol, cellsVol,
                        nodesCellsId, leastSquareCoeff,
                        buildFun, coeffBuild)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
    lag.setScaleValueMat(matOP, row, col, coeff, pbMap,
                         pbMap, 1/dt, size, addRule=True)

    # Diri conditions
    size = mb.ExDiamondDiriFaces(meshDim, row, col, coeff, fBorderD, faces,
                                 facesNormal, diamVol, cellsVol,
                                 buildFun, coeffBuild)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    size = mb.ExDiamondDiriNodes(meshDim, row, coeff, fNotBorder, faces, nodes, cells,
                               facesNormal, diamNormal, diamVol, cellsVol,
                               leastSquareCoeff, funDiri, arrayRho)
    lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)
    size = mb.ExDiamondDiriRhs(meshDim, row, coeff, fBorderD, faces, nodes, cellsCenter,
                               facesNormal, diamNormal, diamVol, cellsVol,
                               diriRhs, (solFun, arrayRho))
    lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)

    ### Build Vol matrix
    size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
    lag.setValueMat(matVol, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    ### All init
    matOP.assemblyBegin()
    matOP.assemblyEnd()

    matVol.assemblyBegin()
    matVol.assemblyEnd()

    if saveMesh:
        msh.saveMeshEData('results/H', lag.getArray(H),
                          cellsGlobId[:nbCellsLoc], pbDim, it, RANK)
    if saveNrm:
        volLoc   = cellsVol[:nbCellsLoc] * (physCells[:nbCellsLoc] == 1000001)
        accuNrmH = []

        nrmH = ut.diff(COMM, lag.getArray(H), lag.getArray(zero), pbDim, 2, volLoc)
        accuNrmH.append((t, nrmH))

    while t + dt <= tEnd:
        t  += dt
        it += 1

        lag.copy(H, prevH)

        ### Linear solver
        lag.solveLin(ksp, matOP, H, 1/dt * matVol * prevH - rhsBd)
        # lag.solveLin(ksp, matOP, H, 1/dt * prevH - rhsBd)

        kspNbIt = ksp.getIterationNumber()
        if kspNbIt == kspMaxIt or kspNbIt == 0:
            raise ValueError("ksp maximum number of iteration reach (%d)" %(kspNbIt))
        if verbose and RANK == 0:
            print('Time %f, Step %d, Iteration %d' %(t, it, kspNbIt))
            sys.stdout.flush()

        if saveMesh:
            msh.saveMeshEData('results/H', lag.getArray(H),
                              cellsGlobId[:nbCellsLoc], pbDim, it, RANK)
        if saveNrm:
            nrmH = ut.diff(COMM, lag.getArray(H), lag.getArray(zero), pbDim, 2, volLoc)
            accuNrmH.append((t, nrmH))

    if saveNrm:
        if RANK == 0:
            if os.path.exists('nrmDir/nrmHInDiam'):
                os.remove('nrmDir/nrmHInDiam')
            fNrmH = open('nrmDir/nrmHInDiam', 'x')
            for cpl in accuNrmH:
                fNrmH.write(str(cpl[0]) + ' ' + str(cpl[1]) + "\n")
            fNrmH.close()
    sys.stdout.flush()

if False:
    if RANK == 0:
        print('---------------------------------------------')
        print('-           Bi-Domaine Split Diamond        -')
        print('---------------------------------------------')

    physRules = {
            100001: 0,
            100002: 1,
            100003: 1,
            }
    notBd   = [-1]
    neuPhys = [100001]
    dirPhys = [100002, 100003]
    neuPhys = []
    dirPhys = [100001, 100002, 100003]
    addNeumLsq(physRules, neuPhys)

    cellsSup = [1000001]
    cellsAir = [1000002]
    physBd    = [100001, 100002]
    physIn    = [(1000001, 1000001)]
    physOut   = [(1000002, 1000002)]
    physInOut = [(1000001, 1000002)]
    physOutIn = [(1000002, 1000001)]
    allCells = [(1000001, 1000001), (1000001, 1000002),
                (1000002, 1000001), (1000002, 1000002)]

    fBorderD = geo.rulesSeclectFaces(faces, dirPhys, allCells,
                                     physCells, nbInnerFaces,
                                     nbSharedFaces,  meshDim)
    fBorderN = geo.rulesSeclectFaces(faces, neuPhys, allCells,
                                     physCells, nbInnerFaces,
                                     nbSharedFaces,  meshDim)
    fNotBorder = geo.rulesSeclectFaces(faces, notBd, allCells,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)

    fIn        = geo.rulesSeclectFaces(faces, notBd, physIn,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fOut       = geo.rulesSeclectFaces(faces, notBd, physOut,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fInOut     = geo.rulesSeclectFaces(faces, notBd, physInOut,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fOutIn     = geo.rulesSeclectFaces(faces, notBd, physOutIn,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)

    # Faces numbering
    nbFacesLoc, facesAllArray, facesMap1 = \
            geo.mappingFacesSubset([fInOut, fOutIn], faces, nbInnerFaces, nbSharedFaces,
                                   allCellsSize, cellsMap1, cellsGlobId,
                                   SIZE, RANK, COMM, meshDim)
    facesMap3 = ut.scaleMap(facesMap1, 3)

    allCells = np.ones(nbCellsLoc, np.int32)
    for i in range(nbCellsLoc):
        allCells[i] = i

    # Space for coeff
    szBuff = 100000000
    row   = np.zeros(szBuff,   np.int32)
    col   = np.zeros(szBuff,   np.int32)
    coeff = np.zeros(szBuff, np.float64)

    t = 0.
    it = 0
    dt = 1e-3
    tEnd = 150*dt

    rhoSup =   1e-3
    rhoAir =   1e-2
    rhoRules = {
            cellsSup[0]: rhoSup,
            cellsAir[0]: rhoAir,
            }

    if False:
        def _fun0(p):
            s = np.zeros(1)
            s[0] = 1.
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            r = rho[rowId]
            vSol = _fun0(pts)
            v = 3 * vD

            row[size] = rowId
            coeff[size] = - r * alpha \
                    * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2]) \
                    * vSol[0] / v
            size += 1

            return 1

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        args):
            fun = args[0]
            rho = args[1][L]

            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   vR * np.dot(nF, nF)
            v += - vA * np.dot(nF, nDA)
            v += - vB * np.dot(nF, nDB)
            v += - vC * np.dot(nF, nDC)
            v  = - v * rho / (3 * vD)

            row[size]    =    L
            coeff[size]  = v[0]
            size        +=    1

            return 1

        buildFun =       mb.lapl3
        idFun    =    mb.cellsId1
        solFun   =          _fun0
        funDiri  =         _diri0
        diriRhs  =    funDiriRhs3
        pbDim    =              1
        pbMap    =      cellsMap1

        kspMaxIt =  5000
        ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs', 'none', 1e-8, 1e-4, kspMaxIt)
    elif False:
        def _fun0(p):
            s = np.zeros(3)
            s[1] = np.cos(p[0])
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            r = rho[rowId]
            vSol = _fun0(pts)
            v = 3 * vD

            row[size]    =  3*rowId
            coeff[size]  = - alpha * r * (nA[1] * nB[1] + nA[2] * nB[2]) * vSol[0] / v
            coeff[size] +=   alpha * r * (nA[1] * nB[0]) * vSol[1] / v
            coeff[size] +=   alpha * r * (nA[2] * nB[0]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +1
            coeff[size]  = - alpha * r * (nA[0] * nB[0] + nA[2] * nB[2]) * vSol[1] / v
            coeff[size] +=   alpha * r * (nA[0] * nB[1]) * vSol[0] / v
            coeff[size] +=   alpha * r * (nA[2] * nB[1]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +2
            coeff[size]  = - alpha * r * (nA[0] * nB[0] + nA[1] * nB[1]) * vSol[2] / v
            coeff[size] +=   alpha * r * (nA[0] * nB[2]) * vSol[0] / v
            coeff[size] +=   alpha * r * (nA[1] * nB[2]) * vSol[1] / v
            size += 1

            return 3

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        args):
            fun = args[0]
            rho = args[1][L]

            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   np.cross(nF, np.cross( nF, vR))
            v += - np.cross(nF, np.cross(nDA, vA))
            v += - np.cross(nF, np.cross(nDB, vB))
            v += - np.cross(nF, np.cross(nDC, vC))
            v  = v * rho / (3 * vD)

            row[size]    =    3*L
            coeff[size]  =   v[0]
            size        +=      1

            row[size]    = 3*L +1
            coeff[size]  =   v[1]
            size        +=      1

            row[size]    = 3*L +2
            coeff[size]  =   v[2]
            size        +=      1

            return 3

        buildFun =   mb.curlCurl3
        idFun    =    mb.cellsId3
        solFun   =          _fun0
        funDiri  =         _diri0
        diriRhs  =    funDiriRhs3
        pbDim    =              3
        pbMap    =      cellsMap3

        kspMaxIt =  5000
        ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs', 'none', 1e-8, 1e-4, kspMaxIt)
    else:
        raise ValueError("No config selected")

    # Cpt Rho
    def computeRho(rho, nbCellsLoc, JLoc, physCells, rhoDict):
        sz = rho.getSizes()[0]
        arrRho = np.zeros(sz, np.float64)
        for i in range(nbCellsLoc):
            v = rhoDict[physCells[i]]
            arrRho[i] = v
        rho.setArray(arrRho)
        return rho

    ### Ghost vec and init
    ghostId = ut.getForeignId(cellsMap1, nbCellsLoc, 1)
    rho = lag.createGhostVec(nbCellsLoc, 1, 1, ghostId)
    computeRho(rho, nbCellsLoc, rho, physCells, rhoRules)
    lag.syncGhostVec(rho)
    arrayRho = lag.getGhostArray(rho)

    coeffBuild = np.zeros(len(faces), np.float64)
    for i in range(len(faces)):
        f = faces[i]
        L = f[3]
        r = arrayRho[L]
        coeffBuild[i] = r / (3 * diamVol[i])
    for lst in [fInOut[0], fInOut[1], fOutIn[0], fOutIn[1]]:
        for i in lst:
            f = faces[i]
            L = f[3]
            R = f[4]
            cL = arrayRho[L] / (3 * splitDVol[i][0])
            cR = arrayRho[R] / (3 * splitDVol[i][1])
            coeffBuild[i] = (cL * cR) / (cL + cR)

    # Vec
    H     = lag.createVec(nbCellsLoc, pbDim, pbDim)
    prevH = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsIn = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsBd = lag.createVec(nbCellsLoc, pbDim, pbDim)
    solH  = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsH  = lag.createVec(nbCellsLoc, pbDim, pbDim)
    zero  = lag.createVec(nbCellsLoc, pbDim, pbDim)

    # Mat
    matOP  = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)
    matVol = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)

    ### Build the matrix
    size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
    lag.setScaleValueMat(matOP, row, col, coeff, pbMap,
                         pbMap, 1/dt, size, addRule=True)

    size = mb.ExDiamond(meshDim, row, col, coeff, fNotBorder, faces,
                        facesNormal, diamNormal, diamVol, cellsVol,
                        nodesCellsId, leastSquareCoeff,
                        buildFun, coeffBuild)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    # Diri conditions
    size = mb.ExDiamondDiriFaces(meshDim, row, col, coeff, fBorderD, faces,
                                 facesNormal, diamVol, cellsVol,
                                 buildFun, coeffBuild)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    size = mb.ExDiamondDiriNodes(meshDim, row, coeff, fOut, faces, nodes, cells,
                                 facesNormal, diamNormal, diamVol, cellsVol,
                                 leastSquareCoeff, funDiri, arrayRho)
    lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)
    size = mb.ExDiamondDiriRhs(meshDim, row, coeff, fBorderD, faces, nodes, cellsCenter,
                               facesNormal, diamNormal, diamVol, cellsVol,
                               diriRhs, (solFun, arrayRho))
    lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)


    ### Build Vol matrix
    size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
    lag.setValueMat(matVol, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    ### All init
    matOP.assemblyBegin()
    matOP.assemblyEnd()

    matVol.assemblyBegin()
    matVol.assemblyEnd()

    if saveMesh:
        msh.saveMeshEData('results/H', lag.getArray(H),
                          cellsGlobId[:nbCellsLoc], pbDim, it, RANK)
    if saveNrm:
        volLoc   = cellsVol[:nbCellsLoc] * (physCells[:nbCellsLoc] == 1000001)
        accuNrmH = []

        nrmH = ut.diff(COMM, lag.getArray(H), lag.getArray(zero), pbDim, 2, volLoc)
        accuNrmH.append((t, nrmH))

    while t + dt <= tEnd:
        t  += dt
        it += 1

        lag.copy(H, prevH)
        lag.copy(1/dt * matVol * prevH - rhsBd, rhsH)

        ### Linear solver
        lag.solveLin(ksp, matOP, H, rhsH)

        kspNbIt = ksp.getIterationNumber()
        if kspNbIt == kspMaxIt or kspNbIt == 0:
            raise ValueError("ksp maximum number of iteration reach (%d)" %(kspNbIt))
        if verbose and RANK == 0:
            print('Time %f, Step %d, Iteration %d' %(t, it, kspNbIt))
            sys.stdout.flush()

        if saveMesh:
            msh.saveMeshEData('results/H', lag.getArray(H),
                              cellsGlobId[:nbCellsLoc], pbDim, it, RANK)
        if saveNrm:
            nrmH = ut.diff(COMM, lag.getArray(H), lag.getArray(zero), pbDim, 2, volLoc)
            accuNrmH.append((t, nrmH))

    if saveNrm:
        if RANK == 0:
            if os.path.exists('nrmDir/nrmHInSplit'):
                os.remove('nrmDir/nrmHInSplit')
            fNrmH = open('nrmDir/nrmHInSplit', 'x')
            for cpl in accuNrmH:
                fNrmH.write(str(cpl[0]) + ' ' + str(cpl[1]) + "\n")
            fNrmH.close()
    sys.stdout.flush()
