import os
import sys

from mpi4py import MPI
# from petsc4py import petsc4py

COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

# opt = PETSc.Options()
# opt['-pc_factor_mat_solver_type'] = 'mumps'

import petsc4py
petsc4py.init(sys.argv)

import numpy as np
import utils as ut
import matBuild as mb
import geometry as geo
import gmshMesh as msh
import petscUtils as lag

from math import sin, pi

if RANK == 0:
    if os.path.exists('results'):
        for file in os.listdir('results'):
            os.remove('results/' + file)
        os.removedirs('results')
        os.mkdir('results')
    else:
        os.mkdir('results')
    files = ['nrmH', 'nrmHSup', 'nrmHDiam', 'nrmHSplit', 'nrmHSmart',
             'pointDiam', 'pointSplit', 'pointSmart', 'nrmHSupra', 'nrmJSupra']
    for f in files:
        if os.path.isfile(f):
            os.remove(f)

verbose  =  True
viewPart = False
saveMesh =  True
saveNrm  =  True

################################################################################
#                                           3D                                 #
################################################################################
if RANK == 0:
    print('---------- 3D ----------')

meshDim = 3
meshFile = 'TODO'
meshFile = '../meshes/2layer3D.msh'
meshFile = '../meshes/adapt2Me.msh'

msh.gmshToHDF5(SIZE, RANK, COMM, meshFile, meshDim)
cells, nodes, cellsGlobId, nodesGlobId,\
            nbCellsLoc, nbNodesLoc, edgesDom,\
            physCells, physEdgesDom = msh.readPartMesh('hdf5Mesh', RANK)
if verbose:
    print("Cells local (%d) size %d and total size %d."
              %(RANK, nbCellsLoc, len(cells)))
    sys.stdout.flush()

# Build basic mesh info
geo.setLocalNodeId(cells, nodesGlobId, edgesDom)
edgesNodesId = geo.getNodesEdges(edgesDom, meshDim)
nodesCellsId = geo.cellsAroundNodes(cells, len(nodes))
faces, nbInnerFaces, nbSharedFaces = geo.facesInfo(cells, nodes,
            edgesDom, physEdgesDom, nbCellsLoc, cellsGlobId, meshDim)
if verbose:
    print("(%d) Number of local faces %d and shared %d."
              %(RANK, nbInnerFaces, nbSharedFaces))
    sys.stdout.flush()

# Build mesh info num methods
cellsCenter = geo.getCellsCenter(cells, nodes)
facesNormal = geo.getFacesNormal(faces, nodes, cellsCenter, meshDim)
cellsVol    = geo.getCellsVol(cells, nodes, meshDim)
leastSquareCoeff = geo.lastSquaresWeight(nodes, nodesCellsId, nbNodesLoc,
            cellsCenter, nbCellsLoc, faces, meshDim)

# Diam info and splitted diam info
diamVol, diamNormal = geo.getInnerDiamond(faces, nodes, cellsCenter, meshDim)
splitDVol, splitDNormal = geo.getSplitDiamond(meshDim, faces, nodes,
            cellsCenter, nbInnerFaces+nbSharedFaces)

# Cells, Nodes and Faces numbering
allCellsSize, cellsMap1 = geo.globToLocNumbering(SIZE, RANK, COMM,
            cellsGlobId, nbCellsLoc)
allNodesSize, nodesMap1 = geo.globToLocNumbering(SIZE, RANK, COMM,
            nodesGlobId, nbNodesLoc)

cellsMap3 = ut.scaleMap(cellsMap1, 3)
nodesMap3 = ut.scaleMap(nodesMap1, 3)

### Get min h value
if verbose:
    hmi, hma = geo.minMaxEdgesLength(COMM, cells, nodes, nbCellsLoc)
    if RANK == 0:
        print("-----------------------------------------")
        print("Min h value: %2.8f" %(hmi))
        print("Max h value: %2.8f" %(hma))
        print("-----------------------------------------")
### View the mesh partition
if viewPart:
    debug = np.zeros(nbCellsLoc, np.float64)
    for i in range(nbCellsLoc):
        debug[i] = RANK
    msh.saveMeshEData('results/part', debug, cellsGlobId, 1, 0, RANK)

def addNeumLsq(rule, neuVal):
    nodesStrongPhys = []

    assert(len(edgesDom) == len(physEdgesDom))
    tmp = []
    for iEdg in range(len(edgesDom)):
        edg = edgesDom[iEdg]
        if edg[0] != -1:
            tmp.append((edg[0], physEdgesDom[iEdg]))
        if edg[1] != -1:
            tmp.append((edg[1], physEdgesDom[iEdg]))
    tmp.sort()

    sz = len(tmp)
    if len(tmp) == 0:
        return
    elif len(tmp) > 2:
        for i in range(sz-1):
            if tmp[i][0] != tmp[i+1][0]:
                nodesStrongPhys.append(tmp[i])
            elif tmp[i][1] != tmp[i+1][1]:
                if physRules[tmp[i][1]] > physRules[tmp[i+1][1]]:
                    tmp[i+1] = (tmp[i][0], tmp[i][1])
        if tmp[-2][0] != tmp[-1][0]:
            nodesStrongPhys.append(tmp[-2])
        elif tmp[-2][1] != tmp[-1][1]:
            if physRules[tmp[-2][1]] > physRules[tmp[-1][1]]:
                tmp[-1] = (tmp[-1][0], tmp[-2][-1])
        nodesStrongPhys.append(tmp[-1])

    nbEdgesNode = len(nodesStrongPhys)
    for i in range(nbEdgesNode):
        if nodesStrongPhys[i][1] in neuVal:
            iNd = nodesStrongPhys[i][0]
            nbC = len(nodesCellsId[iNd])
            if len(leastSquareCoeff[iNd]) != 0:
                continue
            for j in range(nbC):
                leastSquareCoeff[iNd].append(1 / nbC)

if RANK == 0:
    print('---------------------------------------------')
    print('-           Bi-Domaine Linearized           -')
    print('---------------------------------------------')

physRules = {
        100001: 0,
        100002: 1,
        100003: 1,
        }
notBd   = [-1]
neuPhys = [100001]
dirPhys = [100002, 100003]
neuPhys = []
dirPhys = [100001, 100002, 100003]
addNeumLsq(physRules, neuPhys)

cellsSup = [1000001]
cellsAir = [1000002]
physBd    = [100001, 100002]
physIn    = [(1000001, 1000001)]
physOut   = [(1000002, 1000002)]
physInOut = [(1000001, 1000002)]
physOutIn = [(1000002, 1000001)]
allCells = [(1000001, 1000001), (1000001, 1000002),
            (1000002, 1000001), (1000002, 1000002)]

fBorderD = geo.rulesSeclectFaces(faces, dirPhys, allCells,
                                 physCells, nbInnerFaces,
                                 nbSharedFaces,  meshDim)
fBorderN = geo.rulesSeclectFaces(faces, neuPhys, allCells,
                                 physCells, nbInnerFaces,
                                 nbSharedFaces,  meshDim)
fNotBorder = geo.rulesSeclectFaces(faces, notBd, allCells,
                                   physCells, nbInnerFaces,
                                   nbSharedFaces,  meshDim)

fIn        = geo.rulesSeclectFaces(faces, notBd, physIn,
                                   physCells, nbInnerFaces,
                                   nbSharedFaces,  meshDim)
fOut       = geo.rulesSeclectFaces(faces, notBd, physOut,
                                   physCells, nbInnerFaces,
                                   nbSharedFaces,  meshDim)
fInOut     = geo.rulesSeclectFaces(faces, notBd, physInOut,
                                   physCells, nbInnerFaces,
                                   nbSharedFaces,  meshDim)
fOutIn     = geo.rulesSeclectFaces(faces, notBd, physOutIn,
                                   physCells, nbInnerFaces,
                                   nbSharedFaces,  meshDim)

cellsJac = [(1000001, 1000002), (1000002, 1000001), (1000002, 1000002)]
fJac     = geo.rulesSeclectFaces(faces, notBd, allCells,
                                 physCells, nbInnerFaces,
                                 nbSharedFaces,  meshDim)

allCells = np.ones(nbCellsLoc, np.int32)
for i in range(nbCellsLoc):
    allCells[i] = i

# Space for coeff
szBuff = 10000000
row   = np.zeros(szBuff,   np.int32)
col   = np.zeros(szBuff,   np.int32)
coeff = np.zeros(szBuff, np.float64)

t = 0.
it = 0
dt = 1e-6
tEnd = 5e-3

mu0 = 4 * np.pi * 1e-7
Ec  = 1e-4
Jc  =  1e7
n   =   10

rhoSup =   1e-3
rhoAir =   1e-3
rhoRules = {
        cellsSup[0]: (  True, rhoAir),
        cellsAir[0]: ( False, rhoAir),
        }

if True:
    def _fun0(p):
        s = np.zeros(3)
        s[0] = 1e-2 / mu0
        return s
    def _diri0(row, coeff, rowId, size, pts,
               nA, nB, vC, vD, alpha, rho):
        # r = rho[rowId]
        r = rhoAir / mu0
        vSol = _fun0(pts)
        v = 3 * vD

        row[size]    =  3*rowId
        coeff[size]  = - alpha * r * (nA[1] * nB[1] + nA[2] * nB[2]) * vSol[0] / v
        coeff[size] +=   alpha * r * (nA[1] * nB[0]) * vSol[1] / v
        coeff[size] +=   alpha * r * (nA[2] * nB[0]) * vSol[2] / v
        size += 1

        row[size]    = 3*rowId +1
        coeff[size]  = - alpha * r * (nA[0] * nB[0] + nA[2] * nB[2]) * vSol[1] / v
        coeff[size] +=   alpha * r * (nA[0] * nB[1]) * vSol[0] / v
        coeff[size] +=   alpha * r * (nA[2] * nB[1]) * vSol[2] / v
        size += 1

        row[size]    = 3*rowId +2
        coeff[size]  = - alpha * r * (nA[0] * nB[0] + nA[1] * nB[1]) * vSol[2] / v
        coeff[size] +=   alpha * r * (nA[0] * nB[2]) * vSol[0] / v
        coeff[size] +=   alpha * r * (nA[1] * nB[2]) * vSol[1] / v
        size += 1

        return 3

    def funDiriRhs3(row, coeff, L, size,
                    pL, pA, pB, pC, nF, nD, vL, vD,
                    args):
        fun = args[0]
        # rho = args[1][L]
        r = rhoAir / mu0

        vR = fun((2. / 3.) * (pA + pB + pC) - pL)
        vA = fun(pA)
        vB = fun(pB)
        vC = fun(pC)

        nDA = nD[1] + nD[4]
        nDB = nD[2] + nD[5]
        nDC = nD[0] + nD[3]

        v  =   np.cross(nF, np.cross( nF, vR))
        v += - np.cross(nF, np.cross(nDA, vA))
        v += - np.cross(nF, np.cross(nDB, vB))
        v += - np.cross(nF, np.cross(nDC, vC))
        v  = v * r / (3 * vD)

        row[size]    =    3*L
        coeff[size]  =   v[0]
        size        +=      1

        row[size]    = 3*L +1
        coeff[size]  =   v[1]
        size        +=      1

        row[size]    = 3*L +2
        coeff[size]  =   v[2]
        size        +=      1

        return 3

    buildFun =   mb.curlCurl3
    linFun   =      mb.supLin
    jacFun   =      mb.supJac
    ccLin    = mb.curlCurlLin
    idFun    =    mb.cellsId3
    rotFun   =   mb.edgesCurl
    solFun   =          _fun0
    funDiri  =         _diri0
    diriRhs  =    funDiriRhs3
    pbDim    =              3
    pbMap    =      cellsMap3

    kspMaxIt =  1000
    ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
    # ksp = lag.getLinearSolver('fgmres',   'gamg', 1e-8, 1e-4, kspMaxIt)
    # ksp = lag.getLinearSolver(   'bcgs',  'asm', 1e-6, 1e-3, kspMaxIt)

# Cpt Rho
def computeRho(rho, nbCellsLoc, JLoc, physCells, rhoDict):
    sz = rho.getSizes()[0]
    arr = JLoc.getArray()
    arrRho = np.zeros(sz, np.float64)
    for i in range(nbCellsLoc):
        rule = rhoDict[physCells[i]]
        if rule[0]:
            J0   = arr[3*i +0] / Jc
            J1   = arr[3*i +1] / Jc
            J2   = arr[3*i +2] / Jc
            nrmJ = J0*J0 + J1*J1 + J2*J2
            v    = (Ec / (Jc * mu0)) * np.power(nrmJ, (n-1)/2)
        else:
            v    = rule[1] / mu0
        arrRho[i] = v
    rho.setArray(arrRho)
    return rho

### Ghost vec and init
ghostId = ut.getForeignId(cellsMap1, nbCellsLoc, 1)
rho = lag.createGhostVec(nbCellsLoc, 1, 1, ghostId)
arrayRho = lag.getGhostArray(rho)

ghostId = ut.getForeignId(cellsMap1, nbCellsLoc, 3)
J       = lag.createGhostVec(nbCellsLoc, 3, 3, ghostId)
arrayJ  = lag.getGhostArray(J)

for i in range(len(arrayRho)):
    arrayRho[i] = rhoAir / mu0

coeffBuild = np.zeros(     len(faces), np.float64)
coeffLin   = np.zeros((len(faces), 4), np.float64)
coeffMat   = np.zeros((len(faces), 3, 3), np.float64)
coeffJac   = np.zeros((len(faces), 3, 3), np.float64)
for i in range(len(faces)):
    f = faces[i]
    L = f[2]
    r = rhoAir / mu0

    coeffBuild[i] = 0.5 * r / diamVol[i]
    coeffLin[i]   = 0.5 / diamVol[i]
def updateSplit(JLoc):
    for lst in [fInOut[0], fInOut[1], fOutIn[0], fOutIn[1]]:
        for i in lst:
            f = faces[i]
            L = f[2]
            R = f[3]
            cL = 0.5 * arrayRho[L] / splitDVol[i][0]
            cR = 0.5 * arrayRho[R] / splitDVol[i][1]
            coeffBuild[i] = (cL * cR) / (cL + cR)
    for lst in [fIn[0], fIn[1]]:
        for i in lst:
            f = faces[i]
            L = f[2]
            R = f[3]
            cL = 0.5 * arrayRho[L]
            cR = 0.5 * arrayRho[R]
            coeffBuild[i] = 0.5 * (cL + cR) / diamVol[i]
    for lst in [fIn[0], fIn[1]]:
        for i in lst:
            f = faces[i]
            L = f[2]
            R = f[3]

            JL0 = JLoc[3*L +0]
            JL1 = JLoc[3*L +1]
            JL2 = JLoc[3*L +2]

            JR0 = JLoc[3*R +0]
            JR1 = JLoc[3*R +1]
            JR2 = JLoc[3*R +2]

            coeffLin[i][1] = 0.5 * (JL0 + JR0) / Jc
            coeffLin[i][2] = 0.5 * (JL1 + JR1) / Jc
            coeffLin[i][3] = 0.5 * (JL2 + JR2) / Jc
def updateCoeffMat(JLoc):
    for lst in [fInOut[0], fInOut[1], fOutIn[0], fOutIn[1]]:
        for i in lst:
            f = faces[i]
            L = f[2]
            R = f[3]

            matTmp = np.zeros((3, 3), np.float64)

            if physCells[L] == 100001:
                vL = 2 * splitDVol[i][0]
                vR = 2 * splitDVol[i][1]
            else:
                vL = 2 * splitDVol[i][1]
                vR = 2 * splitDVol[i][0]

            JL0 = JLoc[3*L +0]
            JL1 = JLoc[3*L +1]
            JL2 = JLoc[3*L +2]

            JR0 = JLoc[3*R +0]
            JR1 = JLoc[3*R +1]
            JR2 = JLoc[3*R +2]

            Jx = 0.5 * (JL0 + JR0) / Jc
            Jy = 0.5 * (JL1 + JR1) / Jc
            Jz = 0.5 * (JL2 + JR2) / Jc

            nrmJ  = Jx*Jx + Jy*Jy + Jz*Jz
            rhoNL = (Ec / Jc) * np.power(nrmJ, (n-1)/2)

            alpha = rhoNL / vL + rhoAir / vR
            beta  = (n-1) * (Ec / Jc) * np.power(nrmJ, (n-3)/2) / vL

            mb.getInv(alpha, beta, Jx, Jy, Jz, matTmp)

            jacL = np.array([[rhoNL + beta*Jx*Jx,          beta*Jx*Jy,         beta*Jx*Jz],
                            [         beta*Jy*Jx,  rhoNL + beta*Jy*Jy,         beta*Jy*Jz],
                            [         beta*Jz*Jx,          beta*Jz*Jy, rhoNL + beta*Jz*Jz]],
                            np.float64)
            jacR = np.array([[beta*Jx*Jx, beta*Jx*Jy, beta*Jx*Jz],
                             [beta*Jy*Jx, beta*Jy*Jy, beta*Jy*Jz],
                             [beta*Jz*Jx, beta*Jz*Jy, beta*Jz*Jz]],
                            np.float64)

            # print('---------------------------')
            # print(jacL)
            # print(jacR)
            # print(matTmp)
            coeffMat[i][:] = rhoAir * jacL @ matTmp / (vL * vR * mu0)
            coeffJac[i][:] = rhoAir * jacR @ matTmp / (vL * vR * mu0)

# Vec
H     = lag.createVec(nbCellsLoc, pbDim, pbDim)
rhsH  = lag.createVec(nbCellsLoc, pbDim, pbDim)
memH  = lag.createVec(nbCellsLoc, pbDim, pbDim)
tmpH  = lag.createVec(nbCellsLoc, pbDim, pbDim)
prevH = lag.createVec(nbCellsLoc, pbDim, pbDim)
rhsBd = lag.createVec(nbCellsLoc, pbDim, pbDim)
zero  = lag.createVec(nbCellsLoc, pbDim, pbDim)
H.assemblyBegin()
H.assemblyEnd()

# Mat
matOP  = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)
matVol = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)
matRot = lag.createMat((nbCellsLoc, nbCellsLoc), (    3, pbDim), pbDim)
matJac = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)
matIn  = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)

for i in range(nbCellsLoc):
    if physCells[i] == 100001:
        for j in range(pbDim):
            matIn.setValue(pbMap[pbDim * i + j], pbMap[pbDim * i + j], 1., addv=True)
matIn.assemblyBegin()
matIn.assemblyEnd()

### Diri conditions
size = mb.ExDiamondDiriNodes(meshDim, row, coeff, fNotBorder, faces, nodes, cells,
                           facesNormal, diamNormal, diamVol, cellsVol,
                           leastSquareCoeff, funDiri, arrayRho)
lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)
size = mb.ExDiamondDiriRhs(meshDim, row, coeff, fBorderD, faces, nodes, cellsCenter,
                           facesNormal, diamNormal, diamVol, cellsVol,
                           diriRhs, (solFun, arrayRho))
lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)

### Build Vol matrix
size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
lag.setValueMat(matVol, row, col, coeff, pbMap,
                pbMap, size, addRule=True)

### Jac matrix
for i in range(nbCellsLoc):
    matRot.setValue(cellsMap3[3*i +0], pbMap[i], 0., addv=True)
    matRot.setValue(cellsMap3[3*i +1], pbMap[i], 0., addv=True)
    matRot.setValue(cellsMap3[3*i +2], pbMap[i], 0., addv=True)
# Later swap fNorBorder -> fJac
size = mb.CellsEdges(row, col, coeff, fNotBorder, faces,
                     facesNormal, cellsVol, nodesCellsId,
                     leastSquareCoeff, rotFun, meshDim)
lag.setValueMat(matRot, row, col, coeff, cellsMap3, pbMap, size, addRule=True)

### All init
matOP.assemblyBegin()
matOP.assemblyEnd()

matJac.assemblyBegin()
matJac.assemblyEnd()

matVol.assemblyBegin()
matVol.assemblyEnd()

matRot.assemblyBegin()
matRot.assemblyEnd()

if saveMesh:
    matRot.mult(H, J)
    msh.saveMeshEData('results/H', lag.getArray(H),
                      cellsGlobId[:nbCellsLoc], pbDim, it%10, RANK)
    msh.saveMeshEData('results/J', lag.getArray(J) / Jc,
                      cellsGlobId[:nbCellsLoc], 3, it%10, RANK)
    msh.saveMeshEData('results/rho', lag.getArray(rho),
                      cellsGlobId[:nbCellsLoc], 1, it%10, RANK)
if saveNrm:
    volLoc   = cellsVol[:nbCellsLoc] * (physCells[:nbCellsLoc] == 100001)
    accuNrmH = []

    nrmH = ut.diff(COMM, lag.getArray(H), lag.getArray(zero), pbDim, 2, volLoc)
    accuNrmH.append((t, nrmH))

while t + dt <= tEnd:
    it += 1

    lag.copy(H, prevH)

    subIt = 0
    while True:
        matOP.zeroEntries()
        matJac.zeroEntries()
        lag.copy(H, memH)
        # lag.copy(1/dt * matVol * prevH - rhsBd, rhsH)
        lag.copy(1/dt * matVol * prevH - sin(100 * pi * (t + dt)) * rhsBd, rhsH)

        # Diff
        nrmPred = ut.diff(COMM, lag.getArray(memH), lag.getArray(zero),
                          pbDim, 2, np.ones(nbCellsLoc))

        # Update J
        computeRho(rho, nbCellsLoc, J, physCells, rhoRules)
        lag.syncGhostVec(rho)
        lag.syncGhostVec(J)
        arrayJ  = lag.getGhostArray(J)
        updateSplit(arrayJ)
        updateCoeffMat(arrayJ)

        ### Build the matrix
        # matOP
        size = mb.ExDiamond(meshDim, row, col, coeff, fOut, faces,
                            facesNormal, diamNormal, diamVol, cellsVol,
                            nodesCellsId, leastSquareCoeff,
                            buildFun, coeffBuild)
        lag.setValueMat(matOP, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        size = mb.ExDiamond(meshDim, row, col, coeff, fIn, faces,
                            facesNormal, diamNormal, diamVol, cellsVol,
                            nodesCellsId, leastSquareCoeff,
                            linFun, coeffLin)
        lag.setValueMat(matOP, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        size = mb.ExDiamond(meshDim, row, col, coeff, fInOut, faces,
                            facesNormal, diamNormal, diamVol, cellsVol,
                            nodesCellsId, leastSquareCoeff,
                            ccLin, coeffMat)
        lag.setValueMat(matOP, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)
        size = mb.ExDiamond(meshDim, row, col, coeff, fOutIn, faces,
                            facesNormal, diamNormal, diamVol, cellsVol,
                            nodesCellsId, leastSquareCoeff,
                            ccLin, coeffMat)
        lag.setValueMat(matOP, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
        lag.setScaleValueMat(matOP, row, col, coeff, pbMap,
                             pbMap, 1/dt, size, addRule=True)

        # matJac
        size = mb.ExDiamond(meshDim, row, col, coeff, fIn, faces,
                            facesNormal, diamNormal, diamVol, cellsVol,
                            nodesCellsId, leastSquareCoeff,
                            jacFun, coeffLin)
        lag.setValueMat(matJac, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        size = mb.ExDiamond(meshDim, row, col, coeff, fIn, faces,
                            facesNormal, diamNormal, diamVol, cellsVol,
                            nodesCellsId, leastSquareCoeff,
                            ccLin, coeffJac)
        lag.setValueMat(matJac, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        # Diri conditions
        size = mb.ExDiamondDiriFaces(meshDim, row, col, coeff, fBorderD, faces,
                                     facesNormal, diamVol, cellsVol,
                                     buildFun, coeffBuild)
        lag.setValueMat(matOP, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        matOP.assemblyBegin()
        matOP.assemblyEnd()

        matJac.assemblyBegin()
        matJac.assemblyEnd()

        rhsH = rhsH + matIn * matJac * H

        ### Linear solver
        if verbose and RANK == 0:
            print('Solving...')
            sys.stdout.flush()
        lag.solveLin(ksp, matOP, H, rhsH)
        matRot.mult(H, J)
        if verbose and RANK == 0:
            print('Done')
            sys.stdout.flush()

        resi = ksp.getResidualNorm()
        kspNbIt = ksp.getIterationNumber()
        if kspNbIt == kspMaxIt or kspNbIt == 0:
            raise ValueError("ksp maximum number of iteration reach (%d)" %(kspNbIt))

        # Diff
        nrmDiff = ut.diff(COMM, lag.getArray(H), lag.getArray(memH),
                          pbDim, 2, np.ones(nbCellsLoc))
        if verbose and RANK == 0:
            print('Time %f, Step %d, Iteration %d' %(1e7 * (t+dt), it, kspNbIt))
            print('Ratio %f' %(nrmDiff/nrmPred))
            print('Residual', resi)
            sys.stdout.flush()
        if nrmDiff/nrmPred < 1e-3:
            if RANK == 0 and verbose:
                print('---------------------------------------------')
                sys.stdout.flush()
            break
        subIt += 1
        if subIt == 5:
            dt = 0.5 * dt
            subIt = 0
            lag.copy(prevH, H)
            matRot.mult(H, J)
            if verbose and RANK == 0:
                print('Divide timestep: %f' %(1e7 * dt))
                sys.stdout.flush()

    t  += dt
    if subIt <= 2:
        dt = 1.2 * dt
        if verbose and RANK == 0:
            print('Up timestep: %f' %(1e7 * dt))
            sys.stdout.flush()

    if saveMesh:
        msh.saveMeshEData('results/H', lag.getArray(H),
                          cellsGlobId[:nbCellsLoc], pbDim, it%10, RANK)
        msh.saveMeshEData('results/J', lag.getArray(J) / Jc,
                          cellsGlobId[:nbCellsLoc], 3, it%10, RANK)
        msh.saveMeshEData('results/rho', lag.getArray(rho),
                          cellsGlobId[:nbCellsLoc], 1, it%10, RANK)
    if saveNrm:
        nrmH = ut.diff(COMM, lag.getArray(H), lag.getArray(zero), pbDim, 2, volLoc)
        accuNrmH.append((t, nrmH))

if saveNrm:
    if RANK == 0:
        if os.path.exists('nrmDir/nrmHInDiam'):
            os.remove('nrmDir/nrmHInDiam')
        fNrmH = open('nrmDir/nrmHInDiam', 'x')
        for cpl in accuNrmH:
            fNrmH.write(str(cpl[0]) + ' ' + str(cpl[1]) + "\n")
        fNrmH.close()
sys.stdout.flush()
