import os
import sys

from mpi4py import MPI

COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

import numpy as np
import utils as ut
import matBuild as mb
import geometry as geo
import gmshMesh as msh
import petscUtils as lag

if RANK == 0:
    if os.path.exists('results'):
        for file in os.listdir('results'):
            os.remove('results/' + file)
        os.removedirs('results')
        os.mkdir('results')
    else:
        os.mkdir('results')
    files = ['nrmH', 'nrmHSup', 'nrmHDiam', 'nrmHSplit', 'nrmHSmart',
             'pointDiam', 'pointSplit', 'pointSmart', 'nrmHSupra', 'nrmJSupra']
    for f in files:
        if os.path.isfile(f):
            os.remove(f)

verbose  =  True
viewPart = False
saveMesh =  True

################################################################################
#                                           3D                                 #
################################################################################
if RANK == 0:
    print('---------- 3D ----------')

meshDim = 3
meshFile = 'sphere3D.msh'
meshFile = 'cube3D.msh'
meshFile = '../meshes/doubleSphere-1.msh'

msh.gmshToHDF5(SIZE, RANK, COMM, meshFile, meshDim)
cells, nodes, cellsGlobId, nodesGlobId,\
            nbCellsLoc, nbNodesLoc, edgesDom,\
            physCells, physEdgesDom = msh.readPartMesh('hdf5Mesh', RANK)

if verbose:
    print("Cells local (%d) size %d and total size %d."
              %(RANK, nbCellsLoc, len(cells)))
    sys.stdout.flush()

# Build basic mesh info
geo.setLocalNodeId(cells, nodesGlobId, edgesDom)
edgesNodesId = geo.getNodesEdges(edgesDom, meshDim)
nodesCellsId = geo.cellsAroundNodes(cells, len(nodes))
faces, nbInnerFaces, nbSharedFaces = geo.facesInfo(cells, nodes,
            edgesDom, physEdgesDom, nbCellsLoc, cellsGlobId, meshDim)
if verbose:
    print("(%d) Number of local faces %d and shared %d."
              %(RANK, nbInnerFaces, nbSharedFaces))
    sys.stdout.flush()

# Build mesh info num methods
cellsCenter = geo.getCellsCenter(cells, nodes)
facesNormal = geo.getFacesNormal(faces, nodes, cellsCenter, meshDim)
cellsVol    = geo.getCellsVol(cells, nodes, meshDim)
leastSquareCoeff = geo.lastSquaresWeight(nodes, nodesCellsId, nbNodesLoc,
            cellsCenter, nbCellsLoc, faces, meshDim)

# Diam info and splitted diam info
diamVol, diamNormal = geo.getInnerDiamond(faces, nodes, cellsCenter, meshDim)
splitDVol, splitDNormal = geo.getSplitDiamond(meshDim, faces, nodes,
            cellsCenter, nbInnerFaces+nbSharedFaces)

# Cells, Nodes and Faces numbering
allCellsSize, cellsMap1 = geo.globToLocNumbering(SIZE, RANK, COMM,
            cellsGlobId, nbCellsLoc)
allNodesSize, nodesMap1 = geo.globToLocNumbering(SIZE, RANK, COMM,
            nodesGlobId, nbNodesLoc)

cellsMap3 = ut.scaleMap(cellsMap1, 3)
nodesMap3 = ut.scaleMap(nodesMap1, 3)

### Get min h value
if verbose:
    hmi, hma = geo.minMaxEdgesLength(COMM, cells, nodes, nbCellsLoc)
    if RANK == 0:
        print("-----------------------------------------")
        print("Min h value: %2.8f" %(hmi))
        print("Max h value: %2.8f" %(hma))
        print("-----------------------------------------")
### View the mesh partition
if viewPart:
    debug = np.zeros(nbCellsLoc, np.float64)
    for i in range(nbCellsLoc):
        debug[i] = RANK
    msh.saveMeshEData('results/part', debug, cellsGlobId, 1, 0, RANK)

def addNeumLsq(rule, neuVal):
    nodesStrongPhys = []

    assert(len(edgesDom) == len(physEdgesDom))
    tmp = []
    for iEdg in range(len(edgesDom)):
        edg = edgesDom[iEdg]
        if edg[0] != -1:
            tmp.append((edg[0], physEdgesDom[iEdg]))
        if edg[1] != -1:
            tmp.append((edg[1], physEdgesDom[iEdg]))
        if edg[2] != -1:
            tmp.append((edg[2], physEdgesDom[iEdg]))
    tmp.sort()

    sz = len(tmp)
    if len(tmp) == 0:
        return
    elif len(tmp) > 2:
        for i in range(sz-1):
            if tmp[i][0] != tmp[i+1][0]:
                nodesStrongPhys.append(tmp[i])
            elif tmp[i][1] != tmp[i+1][1]:
                if physRules[tmp[i][1]] > physRules[tmp[i+1][1]]:
                    tmp[i+1] = (tmp[i][0], tmp[i][1])
        if tmp[-2][0] != tmp[-1][0]:
            nodesStrongPhys.append(tmp[-2])
        elif tmp[-2][1] != tmp[-1][1]:
            if physRules[tmp[-2][1]] > physRules[tmp[-1][1]]:
                tmp[-1] = (tmp[-1][0], tmp[-2][-1])
        nodesStrongPhys.append(tmp[-1])

    nbEdgesNode = len(nodesStrongPhys)
    for i in range(nbEdgesNode):
        if nodesStrongPhys[i][1] in neuVal:
            iNd = nodesStrongPhys[i][0]
            nbC = len(nodesCellsId[iNd])
            if len(leastSquareCoeff[iNd]) != 0:
                continue
            for j in range(nbC):
                leastSquareCoeff[iNd].append(1 / nbC)
if False:
    physRules = {
            100001: 0,
            100002: 1,
            100003: 1,
            }
    neuPhys = [100001]
    dirPhys = [100002, 100003]
    neuPhys = []
    dirPhys = [100001, 100002, 100003]
    addNeumLsq(physRules, neuPhys)

    allCells = [(1000001, 1000001), (1000002, 1000002),
                (1000001, 1000002), (1000002, 1000001)]
    fBorderD = geo.rulesSeclectFaces(faces, dirPhys, allCells,
                                     physCells, nbInnerFaces,
                                     nbSharedFaces,  meshDim)
    fBorderN = geo.rulesSeclectFaces(faces, neuPhys, allCells,
                                     physCells, nbInnerFaces,
                                     nbSharedFaces,  meshDim)
    fNotBorder = geo.rulesSeclectFaces(faces, [-1], allCells,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)

    allCells = np.ones(nbCellsLoc, np.int32)
    for i in range(nbCellsLoc):
        allCells[i] = i

    coeffBuild = np.zeros(len(faces), np.float64)
    for i in range(len(faces)):
        coeffBuild[i] = 1. / (3. * diamVol[i])

    # Space for coeff
    szBuff = 1000000000
    row   = np.zeros(szBuff,   np.int32)
    col   = np.zeros(szBuff,   np.int32)
    coeff = np.zeros(szBuff, np.float64)

    if False:
        def _fun0(p):
            s = np.zeros(1)
            s[0] = np.cos(p[0])
            return s
        def _rhs0(p):
            s = np.zeros(1)
            s[0] = 2*np.cos(p[0])
            return s
        def _grad(p):
            s = np.zeros(3)
            s[0] = - np.sin(p[0])
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            vSol = _fun0(pts)
            v = 3 * vD

            row[size]   =   rowId
            coeff[size] = - alpha * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2]) * vSol[0] / v
            size += 1

            return 1
        def divGrad(vL, vR, vA, vB, vC, nF, nD, vD, vF):
            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            val  =  (nF[0]*nF[0] + nF[1]*nF[1] + nF[2]*nF[2]) * (vR[0] - vL[0])
            val += -(nF[0]*nDA[0] + nF[1]*nDA[1] + nF[2]*nDA[2]) * vA[0]
            val += -(nF[0]*nDB[0] + nF[1]*nDB[1] + nF[2]*nDB[2]) * vB[0]
            val += -(nF[0]*nDC[0] + nF[1]*nDC[1] + nF[2]*nDC[2]) * vC[0]
            val  = - val / (3 * vD)

            return val

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        fun):
            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   vR * np.dot(nF, nF)
            v += - vA * np.dot(nF, nDA)
            v += - vB * np.dot(nF, nDB)
            v += - vC * np.dot(nF, nDC)
            v  = - v / (3 * vD)

            row[size]    =    L
            coeff[size]  = v[0]
            size        +=    1

            return 1
        def funNeum3(row, coeff, L, size,
                        pA, pB, pC, nF, vL, fun):
            vF = fun(0.5 * (pA + pB))
            v = - np.dot(nF, vF)

            row[size]   = L
            coeff[size] = v
            size += 1

            return 1

        buildFun =       mb.lapl3
        idFun    =    mb.cellsId1
        solFun   =          _fun0
        rhsFun   =          _rhs0
        opFun    =          _grad
        funDiri  =         _diri0
        trueOP   =        divGrad
        diriRhs  =    funDiriRhs3
        neumRhs  =       funNeum3
        pbDim    =              1
        pbMap    =      cellsMap1

        kspMaxIt =  500
        ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs',  'asm', 1e-8, 1e-4, kspMaxIt)
    elif False:
        def _fun0(p):
            s = np.zeros(3)
            s[1] = np.cos(p[0])
            return s
        def _rhs0(p):
            s = np.zeros(3)
            s[1] = 2*np.cos(p[0])
            return s
        def _curl(p):
            s = np.zeros(3)
            s[2] = -np.sin(p[0])
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            vSol = _fun0(pts)
            v = 3 * vD

            row[size]    =  3*rowId
            coeff[size]  = - alpha * (nA[1] * nB[1] + nA[2] * nB[2]) * vSol[0] / v
            coeff[size] +=   alpha * (nA[1] * nB[0]) * vSol[1] / v
            coeff[size] +=   alpha * (nA[2] * nB[0]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +1
            coeff[size]  =  - alpha * (nA[0] * nB[0] + nA[2] * nB[2]) * vSol[1] / v
            coeff[size] +=    alpha * (nA[0] * nB[1]) * vSol[0] / v
            coeff[size] +=    alpha * (nA[2] * nB[1]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +2
            coeff[size]  =  - alpha * (nA[0] * nB[0] + nA[1] * nB[1]) * vSol[2] / v
            coeff[size] +=    alpha * (nA[0] * nB[2]) * vSol[0] / v
            coeff[size] +=    alpha * (nA[1] * nB[2]) * vSol[1] / v
            size += 1

            return 3
        def curlCurl(vL, vR, vA, vB, vC, nF, nD, vD, vF):
            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   np.cross(nF, np.cross( nF, vR))
            v += - np.cross(nF, np.cross(nDA, vA))
            v += - np.cross(nF, np.cross(nDB, vB))
            v += - np.cross(nF, np.cross(nDC, vC))
            v  =   v / (3 * vD)

            return v

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        fun):
            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   np.cross(nF, np.cross( nF, vR))
            v += - np.cross(nF, np.cross(nDA, vA))
            v += - np.cross(nF, np.cross(nDB, vB))
            v += - np.cross(nF, np.cross(nDC, vC))
            v  = v / (3 * vD)

            row[size]    =    3*L
            coeff[size]  =   v[0]
            size        +=      1

            row[size]    = 3*L +1
            coeff[size]  =   v[1]
            size        +=      1

            row[size]    = 3*L +2
            coeff[size]  =   v[2]
            size        +=      1

            return 3
        def funNeum3(row, coeff, L, size,
                        pA, pB, pC, nF, vL, fun):
            vF = fun((pA + pB + pC) / 3.)
            v  = np.cross(nF, vF)

            row[size]   =  3*L +0
            coeff[size] =    v[0]
            size += 1
            row[size]   =  3*L +1
            coeff[size] =    v[1]
            size += 1
            row[size]   =  3*L +2
            coeff[size] =    v[2]
            size += 1

            return 3

        buildFun =   mb.curlCurl3
        idFun    =    mb.cellsId3
        solFun   =          _fun0
        rhsFun   =          _rhs0
        opFun    =          _curl
        funDiri  =         _diri0
        trueOP   =       curlCurl
        diriRhs  =    funDiriRhs3
        neumRhs  =       funNeum3
        pbDim    =              3
        pbMap    =      cellsMap3

        kspMaxIt =  500
        # ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'gamg', 1e-8, 1e-4, kspMaxIt)
        ksp = lag.getLinearSolver(   'bcgs', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs',  'asm', 1e-8, 1e-4, kspMaxIt)
    else:
        raise ValueError("No config selected")

    # Vec
    H     = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsIn = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsBd = lag.createVec(nbCellsLoc, pbDim, pbDim)
    solH  = lag.createVec(nbCellsLoc, pbDim, pbDim)
    zero  = lag.createVec(nbCellsLoc, pbDim, pbDim)

    # Mat
    matOP = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)

    # Build the matrix
    size = mb.ExDiamond(meshDim, row, col, coeff, fNotBorder, faces,
                        facesNormal, diamNormal, diamVol, cellsVol,
                        nodesCellsId, leastSquareCoeff,
                        buildFun, coeffBuild)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    # Diri conditions
    sz = mb.ExDiamondDiriNodes(meshDim, row, coeff, fNotBorder, faces, nodes, cells,
                               facesNormal, diamNormal, diamVol, cellsVol,
                               leastSquareCoeff, funDiri, None)
    lag.setValueVec(rhsBd, coeff, row, pbMap, sz, addRule=True)

    if True:
        # Neu and Diri
        size = mb.ExDiamondDiriFaces(meshDim, row, col, coeff, fBorderD, faces,
                                     facesNormal, diamVol, cellsVol,
                                     buildFun, coeffBuild)
        lag.setValueMat(matOP, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        size = mb.ExDiamondDiriRhs(meshDim, row, coeff, fBorderD, faces, nodes, cellsCenter,
                                   facesNormal, diamNormal, diamVol, cellsVol,
                                   diriRhs, solFun)
        lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)

        size = mb.ExDiamondNeum(meshDim, row, coeff, fBorderN, faces, nodes,
                                facesNormal, cellsVol, neumRhs, opFun)
        lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)
    else:
        raise ValueError("No config selected")

    ### All init
    sz = ut.cptVecVal(coeff, row, rhsFun, cellsCenter,
                      cellsVol, nbCellsLoc, pbDim)
    lag.setValueVec(rhsIn, coeff, row, pbMap, sz)
    sz = ut.cptVecVal(coeff, row, solFun, cellsCenter,
                     [1.], nbCellsLoc, pbDim)
    lag.setValueVec(solH, coeff, row, pbMap, sz)

    matOP.assemblyBegin()
    matOP.assemblyEnd()

    ### Linear solver
    lag.solveLin(ksp, matOP, H, rhsIn - rhsBd)
    kspNbIt = ksp.getIterationNumber()
    if kspNbIt == kspMaxIt:
        raise ValueError("ksp maximum number of iteration reach (%d)" %(kspNbIt))
    if verbose and RANK == 0:
        print('Nb itaration: ', kspNbIt)

    volLoc = cellsVol[:nbCellsLoc]
    nrmH = ut.diff(COMM, lag.getArray(solH), lag.getArray(zero),
                   pbDim, 2, volLoc)
    errH = ut.diff(COMM, lag.getArray(solH), lag.getArray(H),
                   pbDim, 2, volLoc)
    if verbose and RANK == 0:
        print('Nrm: %2.11f, Err: %2.11f' %(nrmH, errH))

    if saveMesh:
        msh.saveMeshEData('results/H', lag.getArray(H),
                          cellsGlobId[:nbCellsLoc], pbDim, 0, RANK)
        msh.saveMeshEData('results/solH', lag.getArray(solH),
                          cellsGlobId[:nbCellsLoc], pbDim, 0, RANK)
        msh.saveMeshEData('results/errH', lag.getArray(H) - lag.getArray(solH),
                          cellsGlobId[:nbCellsLoc], pbDim, 0, RANK)

if False:
    physRules = {
            100001: 0,
            100002: 1,
            100003: 1,
            }
    notBd   = [-1]
    neuPhys = [100001]
    dirPhys = [100002, 100003]
    neuPhys = []
    dirPhys = [100001, 100002, 100003]
    addNeumLsq(physRules, neuPhys)

    physIn    = [(1000001, 1000001)]
    physOut   = [(1000002, 1000002)]
    physInOut = [(1000001, 1000002)]
    physOutIn = [(1000002, 1000001)]
    allCells  = [(1000001, 1000001), (1000001, 1000002),
                 (1000002, 1000001), (1000002, 1000002)]

    fBorderD   = geo.rulesSeclectFaces(faces, dirPhys, allCells,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fBorderN   = geo.rulesSeclectFaces(faces, neuPhys, allCells,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fNotBorder = geo.rulesSeclectFaces(faces, notBd, allCells,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)

    fIn        = geo.rulesSeclectFaces(faces, notBd, physIn,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fOut       = geo.rulesSeclectFaces(faces, notBd, physOut,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fInOut     = geo.rulesSeclectFaces(faces, notBd, physInOut,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)
    fOutIn     = geo.rulesSeclectFaces(faces, notBd, physOutIn,
                                       physCells, nbInnerFaces,
                                       nbSharedFaces,  meshDim)

    # Faces numbering
    nbFacesLoc, facesAllArray, facesMap1 = \
            geo.mappingFacesSubset([fInOut, fOutIn], faces, nbInnerFaces, nbSharedFaces,
                                   allCellsSize, cellsMap1, cellsGlobId,
                                   SIZE, RANK, COMM, meshDim)
    facesMap3 = ut.scaleMap(facesMap1, 3)

    allCells = np.ones(nbCellsLoc, np.int32)
    for i in range(nbCellsLoc):
        allCells[i] = i

    coeffBuild = np.zeros(len(faces), np.float64)
    for i in range(len(faces)):
        coeffBuild[i] = 1. / (3. * diamVol[i])
    for lst in [fInOut[0], fInOut[1], fOutIn[0], fOutIn[1]]:
        for i in lst:
            cL = 1. / (3. * splitDVol[i][0])
            cR = 1. / (3. * splitDVol[i][1])
            coeffBuild[i] = (cL * cR) / (cL + cR)

    # Space for coeff
    szBuff = 1000000000
    row   = np.zeros(szBuff,   np.int32)
    col   = np.zeros(szBuff,   np.int32)
    coeff = np.zeros(szBuff, np.float64)

    if False:
        def _fun0(p):
            s = np.zeros(1)
            s[0] = np.cos(p[0])
            return s
        def _rhs0(p):
            s = np.zeros(1)
            s[0] = 2*np.cos(p[0])
            return s
        def _grad(p):
            s = np.zeros(3)
            s[0] = - np.sin(p[0])
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            vSol = _fun0(pts)
            v = 3 * vD

            row[size]   =   rowId
            coeff[size] = - alpha * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2]) * vSol[0] / v
            size += 1

            return 1
        def divGrad(vL, vR, vA, vB, vC, nF, nD, vD, vF):
            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            val  =  (nF[0]*nF[0] + nF[1]*nF[1] + nF[2]*nF[2]) * (vR[0] - vL[0])
            val += -(nF[0]*nDA[0] + nF[1]*nDA[1] + nF[2]*nDA[2]) * vA[0]
            val += -(nF[0]*nDB[0] + nF[1]*nDB[1] + nF[2]*nDB[2]) * vB[0]
            val += -(nF[0]*nDC[0] + nF[1]*nDC[1] + nF[2]*nDC[2]) * vC[0]
            val  = - val / (3 * vD)

            return val

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        fun):
            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   vR * np.dot(nF, nF)
            v += - vA * np.dot(nF, nDA)
            v += - vB * np.dot(nF, nDB)
            v += - vC * np.dot(nF, nDC)
            v  = - v / (3 * vD)

            row[size]    =    L
            coeff[size]  = v[0]
            size        +=    1

            return 1
        def funNeum3(row, coeff, L, size,
                        pA, pB, pC, nF, vL, fun):
            vF = fun(0.5 * (pA + pB))
            v = - np.dot(nF, vF)

            row[size]   = L
            coeff[size] = v
            size += 1

            return 1

        buildFun =       mb.lapl3
        equalFun =       mb.lapl3
        idFun    =    mb.cellsId1
        solFun   =          _fun0
        rhsFun   =          _rhs0
        opFun    =          _grad
        funDiri  =         _diri0
        trueOP   =        divGrad
        diriRhs  =    funDiriRhs3
        neumRhs  =       funNeum3
        pbDim    =              1
        pbMap    =      cellsMap1
        pbFap    =      facesMap1

        kspMaxIt =  500
        ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver( 'fgmres', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs', 'gamg', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs',  'asm', 1e-8, 1e-4, kspMaxIt)
        # ksp = lag.getLinearSolver(   'bcgs', 'none', 1e-8, 1e-4, kspMaxIt)
    elif False:
        def _fun0(p):
            s = np.zeros(3)
            s[1] = np.cos(p[0])
            return s
        def _rhs0(p):
            s = np.zeros(3)
            s[1] = 2*np.cos(p[0])
            return s
        def _curl(p):
            s = np.zeros(3)
            s[2] = -np.sin(p[0])
            return s
        def _diri0(row, coeff, rowId, size, pts,
                   nA, nB, vC, vD, alpha, rho):
            vSol = _fun0(pts)
            v = 3 * vD

            row[size]    =  3*rowId
            coeff[size]  = - alpha * (nA[1] * nB[1] + nA[2] * nB[2]) * vSol[0] / v
            coeff[size] +=   alpha * (nA[1] * nB[0]) * vSol[1] / v
            coeff[size] +=   alpha * (nA[2] * nB[0]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +1
            coeff[size]  =  - alpha * (nA[0] * nB[0] + nA[2] * nB[2]) * vSol[1] / v
            coeff[size] +=    alpha * (nA[0] * nB[1]) * vSol[0] / v
            coeff[size] +=    alpha * (nA[2] * nB[1]) * vSol[2] / v
            size += 1

            row[size]    = 3*rowId +2
            coeff[size]  =  - alpha * (nA[0] * nB[0] + nA[1] * nB[1]) * vSol[2] / v
            coeff[size] +=    alpha * (nA[0] * nB[2]) * vSol[0] / v
            coeff[size] +=    alpha * (nA[1] * nB[2]) * vSol[1] / v
            size += 1

            return 3
        def curlCurl(vL, vR, vA, vB, vC, nF, nD, vD, vF):
            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   np.cross(nF, np.cross( nF, vR))
            v += - np.cross(nF, np.cross(nDA, vA))
            v += - np.cross(nF, np.cross(nDB, vB))
            v += - np.cross(nF, np.cross(nDC, vC))
            v  =   v / (3 * vD)

            return v

        def funDiriRhs3(row, coeff, L, size,
                        pL, pA, pB, pC, nF, nD, vL, vD,
                        fun):
            vR = fun((2. / 3.) * (pA + pB + pC) - pL)
            vA = fun(pA)
            vB = fun(pB)
            vC = fun(pC)

            nDA = nD[1] + nD[4]
            nDB = nD[2] + nD[5]
            nDC = nD[0] + nD[3]

            v  =   np.cross(nF, np.cross( nF, vR))
            v += - np.cross(nF, np.cross(nDA, vA))
            v += - np.cross(nF, np.cross(nDB, vB))
            v += - np.cross(nF, np.cross(nDC, vC))
            v  = v / (3 * vD)

            row[size]    =    3*L
            coeff[size]  =   v[0]
            size        +=      1

            row[size]    = 3*L +1
            coeff[size]  =   v[1]
            size        +=      1

            row[size]    = 3*L +2
            coeff[size]  =   v[2]
            size        +=      1

            return 3
        def funNeum3(row, coeff, L, size,
                        pA, pB, pC, nF, vL, fun):
            vF = fun((pA + pB + pC) / 3.)
            v  = np.cross(nF, vF)

            row[size]   =  3*L +0
            coeff[size] =    v[0]
            size += 1
            row[size]   =  3*L +1
            coeff[size] =    v[1]
            size += 1
            row[size]   =  3*L +2
            coeff[size] =    v[2]
            size += 1

            return 3

        buildFun =   mb.curlCurl3
        equalFun =   mb.curlCurl3
        idFun    =    mb.cellsId3
        solFun   =          _fun0
        rhsFun   =          _rhs0
        opFun    =          _curl
        funDiri  =         _diri0
        trueOP   =       curlCurl
        diriRhs  =    funDiriRhs3
        neumRhs  =       funNeum3
        pbDim    =              3
        pbMap    =      cellsMap3
        pbFap    =      facesMap3

        kspMaxIt =  5000
        # ksp = lag.getLinearSolver('preonly',   'lu', 1e-8, 1e-8, kspMaxIt)
        ksp = lag.getLinearSolver( 'fgmres', 'none', 1e-8, 1e-4, kspMaxIt)
        ksp = lag.getLinearSolver(   'bcgs', 'none', 1e-8, 1e-4, kspMaxIt)
    else:
        raise ValueError("No config selected")

    # Vec
    H     = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsIn = lag.createVec(nbCellsLoc, pbDim, pbDim)
    rhsBd = lag.createVec(nbCellsLoc, pbDim, pbDim)
    solH  = lag.createVec(nbCellsLoc, pbDim, pbDim)
    zero  = lag.createVec(nbCellsLoc, pbDim, pbDim)

    # Mat
    matOP = lag.createMat((nbCellsLoc, nbCellsLoc), (pbDim, pbDim), pbDim)

    ### Build the matrix
    size = mb.ExCells(row, col, coeff, allCells, cellsVol, idFun)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    size = mb.ExDiamond(meshDim, row, col, coeff, fNotBorder, faces,
                        facesNormal, diamNormal, diamVol, cellsVol,
                        nodesCellsId, leastSquareCoeff,
                        buildFun, coeffBuild)
    lag.setValueMat(matOP, row, col, coeff, pbMap,
                    pbMap, size, addRule=True)

    # Diri conditions
    sz = mb.ExDiamondDiriNodes(meshDim, row, coeff, fOut, faces, nodes, cells,
                               facesNormal, diamNormal, diamVol, cellsVol,
                               leastSquareCoeff, funDiri, None)
    lag.setValueVec(rhsBd, coeff, row, pbMap, sz, addRule=True)

    if True:
        # Neu and Diri
        size = mb.ExDiamondDiriFaces(meshDim, row, col, coeff, fBorderD, faces,
                                     facesNormal, diamVol, cellsVol,
                                     buildFun, coeffBuild)
        lag.setValueMat(matOP, row, col, coeff, pbMap,
                        pbMap, size, addRule=True)

        size = mb.ExDiamondDiriRhs(meshDim, row, coeff, fBorderD, faces, nodes, cellsCenter,
                                   facesNormal, diamNormal, diamVol, cellsVol,
                                   diriRhs, solFun)
        lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)

        size = mb.ExDiamondNeum(meshDim, row, coeff, fBorderN, faces, nodes,
                                facesNormal, cellsVol, neumRhs, opFun)
        lag.setValueVec(rhsBd, coeff, row, pbMap, size, addRule=True)
    else:
        raise ValueError("No config selected")

    ### All init
    sz = ut.cptVecVal(coeff, row, rhsFun, cellsCenter,
                      cellsVol, nbCellsLoc, pbDim)
    lag.setValueVec(rhsIn, coeff, row, pbMap, sz)
    sz = ut.cptVecVal(coeff, row, solFun, cellsCenter,
                      [1.], nbCellsLoc, pbDim)
    lag.setValueVec(solH, coeff, row, pbMap, sz)

    matOP.assemblyBegin()
    matOP.assemblyEnd()

    ### Linear solver
    lag.solveLin(ksp, matOP, H, rhsIn - rhsBd)
    kspNbIt = ksp.getIterationNumber()
    if kspNbIt == kspMaxIt:
        raise ValueError("ksp maximum number of iteration reach (%d)" %(kspNbIt))
    if verbose and RANK == 0:
        print('Nb itaration: ', kspNbIt)

    volLoc = cellsVol[:nbCellsLoc]
    nrmH = ut.diff(COMM, lag.getArray(solH), lag.getArray(zero),
                   pbDim, 2, volLoc)
    errH = ut.diff(COMM, lag.getArray(solH), lag.getArray(H),
                   pbDim, 2, volLoc)
    if verbose and RANK == 0:
        print('Nrm: %2.11f, Err: %2.11f' %(nrmH, errH))

    if saveMesh:
        msh.saveMeshEData('results/H', lag.getArray(H),
                          cellsGlobId[:nbCellsLoc], pbDim, 0, RANK)
        msh.saveMeshEData('results/solH', lag.getArray(solH),
                          cellsGlobId[:nbCellsLoc], pbDim, 0, RANK)
        msh.saveMeshEData('results/errH', lag.getArray(H) - lag.getArray(solH),
                          cellsGlobId[:nbCellsLoc], pbDim, 0, RANK)
