import numpy as np

from mpi4py import MPI

################################################################################
#                                           Basic mesh building                #
################################################################################

# For all nodeId read from hdf5, change the global id to a local one.
def setLocalNodeId(cells, nodesGlobId, edges):
    lookup = []
    sz = len(nodesGlobId)
    for i in range(sz):
        lookup.append((nodesGlobId[i], i))
    lookup.sort()

    # Return lookup[i][1] when lookup[i][0] is provided (binary search).
    def getId(lookup, idx, sz):
        begin = 0
        end = sz-1
        while begin < end:
            mid = int((begin + end)/2)
            if lookup[mid][0] == idx:
                return lookup[mid][1]
            elif lookup[mid][0] < idx:
                begin = mid+1
            else:
                end = mid
        assert(begin == end)
        if idx != lookup[begin][0]:
            return -1
        return lookup[begin][1]

    for i in range(len(cells)):
        tmp = cells[i]
        for j in range(len(tmp)):
            tmp[j] = getId(lookup, tmp[j], sz)

    for i in range(len(edges)):
        tmp = edges[i]
        for j in range(len(tmp)):
            tmp[j] = getId(lookup, tmp[j], sz)

# nodesCellsId[i][j] if the j-th cells with nodes[i] as nodes.
def cellsAroundNodes(cells, nbCells):
    nodesCellsId = [[] for i in range(nbCells)]
    for i in range(len(cells)):
        for j in range(len(cells[i])):
            nodesCellsId[cells[i][j]].append(i)
    return nodesCellsId

# Make the difference between 2D and 3D
def facesInfo(cells, nodes, edges, physEdges,
              nbCellsLoc, cellsGlobId, meshDim):
    if meshDim == 2:
        return facesInfo2D(cells, nodes, edges, physEdges,
                           nbCellsLoc, cellsGlobId)
    elif meshDim == 3:
        return facesInfo3D(cells, nodes, edges, physEdges,
                           nbCellsLoc, cellsGlobId)
    else:
        raise ValueError("Invalid mesh dimension")

# Build faces array for each faces (of 2D mesh):
# faces description:
#  - id nA, id nB, id cL, id cR, phys, sens
#  - rule id nA < id nB
#         id cL < id cR
#  - sens of original global id (for faces normales)
def facesInfo2D(cells, nodes, edges, physEdges, nbCellsLoc, cellsGlobId):
    facesData = []
    facesSorted = []

    # Add all faces encountered (even duplicates)
    for i in range(len(cells)):
        n0 = cells[i][0]
        n1 = cells[i][1]
        n2 = cells[i][2]

        if n0 < n1:
            facesData.append([n0, n1, i])
        else:
            facesData.append([n1, n0, i])
        if n0 < n2:
            facesData.append([n0, n2, i])
        else:
            facesData.append([n2, n0, i])
        if n1 < n2:
            facesData.append([n1, n2, i])
        else:
            facesData.append([n2, n1, i])
    facesData.sort()

    # Merge duplicate to create face data
    sz = len(facesData)
    i = 1
    while i < sz:
        nA = facesData[i-1][0]
        nB = facesData[i-1][1]
        if nA == facesData[i][0] and nB == facesData[i][1]:
            facesSorted.append([nA, nB, facesData[i-1][2], facesData[i][2], -1, 1])
            i += 2
        else:
            facesSorted.append([nA, nB, facesData[i-1][2], -1, -1, 1])
            i += 1

    # Last facesData case
    nA = facesData[sz-1][0]
    nB = facesData[sz-1][1]
    if not(nA == facesSorted[-1][0] and nB == facesSorted[-1][1]):
        facesSorted.append([nA, nB, facesData[sz-1][2], -1, -1, 1])

    # Return edges index from nA and nB (binary search)
    def getId(faces, nA, nB, sz):
        begin = 0
        end = sz-1
        while begin < end:
            mid = int((begin + end)/2)
            if faces[mid][0] == nA and faces[mid][1] == nB:
                return mid
            elif faces[mid][0] < nA:
                begin = mid+1
            elif faces[mid][0] == nA and faces[mid][1] < nB:
                begin = mid+1
            else:
                end = mid
        assert(begin == end)
        return begin
    # Add edges tag from domian border
    for i in range(len(edges)):
        if edges[i][0] < edges[i][1]:
            nA = edges[i][0]
            nB = edges[i][1]
        else:
            nA = edges[i][1]
            nB = edges[i][0]

        pos = getId(facesSorted, nA, nB, len(facesSorted))

        # Check that the position returned correspond to the edges
        # before change it's physical property
        gA = (nA == facesSorted[pos][0])
        gB = (nB == facesSorted[pos][1])
        if gA and gB:
            facesSorted[pos][4] = physEdges[i]

    faces, nbFacesInner, nbFacesShared = sortFaces(facesSorted, nbCellsLoc, 2)

    # Correct sens with global id
    # Not for innerfaces ?
    for i in range(nbFacesInner, len(faces)):
        if faces[i][3] >= nbCellsLoc:
            glob2 = cellsGlobId[faces[i][2]]
            glob3 = cellsGlobId[faces[i][3]]
            if glob3 < glob2:
                faces[i][5] = -1

    return faces, nbFacesInner, nbFacesShared

# Build faces array for each faces (of 3D mesh):
# faces description:
#  - id nA, id nB, id nC, id cL, id cR, phys, sens
#  - rule id nA < id nB
#         id cL < id cR
#  - sens of original global id (for faces normales)
def facesInfo3D(cells, nodes, edges, physEdges, nbCellsLoc, cellsGlobId):
    facesData = []
    facesSorted = []

    # Add all faces encountered (even duplicates)
    for i in range(len(cells)):
        n = [cells[i][0], cells[i][1],
             cells[i][2], cells[i][3]]
        n.sort()

        n0 = n[0]
        n1 = n[1]
        n2 = n[2]
        n3 = n[3]

        facesData.append([n0, n1, n2, i])
        facesData.append([n0, n1, n3, i])
        facesData.append([n0, n2, n3, i])
        facesData.append([n1, n2, n3, i])
    facesData.sort()

    # Merge duplicate to create face data
    sz = len(facesData)
    i = 1
    while i < sz:
        nA = facesData[i-1][0]
        nB = facesData[i-1][1]
        nC = facesData[i-1][2]
        if nA == facesData[i][0] and nB == facesData[i][1] and nC == facesData[i][2]:
            facesSorted.append([nA, nB, nC, facesData[i-1][3], facesData[i][3], -1, 1])
            i += 2
        else:
            facesSorted.append([nA, nB, nC, facesData[i-1][3], -1, -1, 1])
            i += 1

    # Last facesData case
    nA = facesData[sz-1][0]
    nB = facesData[sz-1][1]
    nC = facesData[sz-1][2]
    if not(nA == facesSorted[-1][0] and nB == facesSorted[-1][1] and nC == facesSorted[-1][2]):
        facesSorted.append([nA, nB, nC, facesData[sz-1][3], -1, -1, 1])

    # Return edges index from nA, nB and nC (binary search)
    def getId(faces, nA, nB, nC, sz):
        begin = 0
        end = sz-1
        while begin < end:
            mid = int((begin + end)/2)
            if faces[mid][0] == nA and faces[mid][1] == nB and faces[mid][2] == nC:
                return mid
            elif faces[mid][0] < nA:
                begin = mid+1
            elif faces[mid][0] == nA and faces[mid][1] < nB:
                begin = mid+1
            elif faces[mid][0] == nA and faces[mid][1] == nB and faces[mid][2] < nC:
                begin = mid+1
            else:
                end = mid
        assert(begin == end)
        return begin

    # Add edges tag from domian border
    for i in range(len(edges)):
        n = [edges[i][0], edges[i][1], edges[i][2]]
        n.sort()

        nA = n[0]
        nB = n[1]
        nC = n[2]

        pos = getId(facesSorted, nA, nB, nC, len(facesSorted))

        # Check that the position returned correspond to the edges
        # before change it's physical property
        gA = (nA == facesSorted[pos][0])
        gB = (nB == facesSorted[pos][1])
        gC = (nC == facesSorted[pos][2])
        if gA and gB and gC:
            facesSorted[pos][5] = physEdges[i]

    faces, nbFacesInner, nbFacesShared = sortFaces(facesSorted, nbCellsLoc, 3)

    # Correct sens with global id
    # Not for innerfaces ?
    for i in range(nbFacesInner, len(faces)):
        if faces[i][4] >= nbCellsLoc:
            cL = cellsGlobId[faces[i][3]]
            cR = cellsGlobId[faces[i][4]]
            if cR < cL:
                faces[i][6] = -1

    return faces, nbFacesInner, nbFacesShared

# Sort the faces to have in the following order:
#  - innerfaces (Left and Right cells are owned)
#  - shared faces (Left owned, Right borrowed)
#  - Borrowed faces (Left and Right borrowed)
# We also put domain edges at the begin of each categorie.
def sortFaces(faces, nbCellsLoc, meshDim):
    if meshDim == 2:
        return sortFaces2D(faces, nbCellsLoc)
    elif meshDim == 3:
        return sortFaces3D(faces, nbCellsLoc)
    else:
        raise ValueError("Invalid mesh dimension")
def sortFaces2D(faces, nbCellsLoc):
    innerBorder = []
    innerFaces = []
    sharedFaces = []
    borrowBorder = []
    borrowFaces = []
    sortedFaces = []

    for i in range(len(faces)):
        if faces[i][4] != -1:
            if faces[i][2] < nbCellsLoc:
                innerBorder.append(faces[i])
            else:
                borrowBorder.append(faces[i])
        else:
            if faces[i][2] >= nbCellsLoc:
                borrowFaces.append(faces[i])
            elif faces[i][3] >= nbCellsLoc:
                sharedFaces.append(faces[i])
            else:
                innerFaces.append(faces[i])

    for lst in [ innerBorder, innerFaces, sharedFaces,
                borrowBorder, borrowFaces]:
        for i in range(len(lst)):
            sortedFaces.append(lst[i])

    return sortedFaces, len(innerBorder) + len(innerFaces), len(sharedFaces)
def sortFaces3D(faces, nbCellsLoc):
    innerBorder = []
    innerFaces = []
    sharedFaces = []
    borrowBorder = []
    borrowFaces = []
    sortedFaces = []

    for i in range(len(faces)):
        if faces[i][5] != -1:
            if faces[i][3] < nbCellsLoc:
                innerBorder.append(faces[i])
            else:
                borrowBorder.append(faces[i])
        else:
            if faces[i][3] >= nbCellsLoc:
                borrowFaces.append(faces[i])
            elif faces[i][4] >= nbCellsLoc:
                sharedFaces.append(faces[i])
            else:
                innerFaces.append(faces[i])

    for lst in [ innerBorder, innerFaces, sharedFaces,
                borrowBorder, borrowFaces]:
        for i in range(len(lst)):
            sortedFaces.append(lst[i])

    return sortedFaces, len(innerBorder) + len(innerFaces), len(sharedFaces)

# Get id of edge nodes
def getNodesEdges(edges, meshDim):
    tmp = []
    edgN = []

    # Get all edge nodes in an array
    if meshDim == 2:
        for edg in edges:
            if edg[0] != -1:
                tmp.append(edg[0])
            if edg[1] != -1:
                tmp.append(edg[1])
    elif meshDim == 3:
        for edg in edges:
            if edg[0] != -1:
                tmp.append(edg[0])
            if edg[1] != -1:
                tmp.append(edg[1])
            if edg[2] != -1:
                tmp.append(edg[2])
    else:
        raise ValueError("Invalid mesh dimension")
    tmp.sort()

    # Remove duplicate
    sz = len(tmp)
    if sz == 0:
        return np.array([], np.int32)

    for i in range(sz-1):
        if tmp[i] != tmp[i+1]:
            edgN.append(tmp[i])
    edgN.append(tmp[sz-1])

    return np.array(edgN, np.int32)
################################################################################
#                                           Compute value for NM scheme        #
################################################################################

# Compute cells center (array).
def getCellsCenter(cells, nodes):
    sz = len(cells)
    center = np.zeros((sz, 3), dtype=np.float64)
    for i in range(sz):
        nbPts = len(cells[i])
        for j in range(nbPts):
            center[i] += nodes[cells[i][j]]
        center[i] /= nbPts

    return center

# Compute cells volume (array).
def getCellsVol(cells, nodes, meshDim):
    if meshDim == 2:
        return getCellsVol2D(cells, nodes)
    elif meshDim == 3:
        return getCellsVol3D(cells, nodes)
    else:
        raise ValueError("Invalid mesh dimension")
def getCellsVol2D(cells, nodes):
    sz = len(cells)
    vol = np.zeros(sz, dtype=np.float64)

    for i in range(sz):
        x0 = nodes[cells[i][0]][0]
        x1 = nodes[cells[i][1]][0]
        x2 = nodes[cells[i][2]][0]

        y0 = nodes[cells[i][0]][1]
        y1 = nodes[cells[i][1]][1]
        y2 = nodes[cells[i][2]][1]

        vol[i] = 1./2. * abs((x0-x1)*(y0-y2) - (x0-x2)*(y0-y1))
    return vol
def getCellsVol3D(cells, nodes):
    sz  = len(cells)
    u   = np.zeros( 3, dtype=np.float64)
    v   = np.zeros( 3, dtype=np.float64)
    w   = np.zeros( 3, dtype=np.float64)
    vol = np.zeros(sz, dtype=np.float64)

    for i in range(sz):
        n0 = cells[i][0]
        n1 = cells[i][1]
        n2 = cells[i][2]
        n3 = cells[i][3]

        # u and v border of faces
        # Last points - other faces center
        u[:] = nodes[n1][:] - nodes[n0][:]
        v[:] = nodes[n2][:] - nodes[n0][:]
        w[:] = nodes[n0][:] + nodes[n1][:] + nodes[n2][:]
        w[:] = w[:] / 3 - nodes[n3][:]

        tmp  = np.cross(u, v) / 2
        vTmp = (w[0]*tmp[0] + w[1]*tmp[1] + w[2]*tmp[2]) / 3
        if vTmp > 0:
            vol[i] =  vTmp
        else:
            vol[i] = -vTmp

    return vol

# Nodes get the tags of their faces (maximum ones).
def getNodesFacesTags(faces, id1stBorrowFaces, nbNodes):
    sz = len(faces)
    tags = -np.ones(nbNodes, dtype=np.int32)

    for i in range(sz):
        if faces[i][4] != -1:
            tags[faces[i][0]] = faces[i][4]
            tags[faces[i][1]] = faces[i][4]
        else:
            break
    return tags

# Cells get the tags of the nodes around them (maximum ones).
def getCellsFacesTags(nodesFTags, nodesCellsId, nbCells):
    sz = len(nodesFTags)
    tags = -np.ones(nbCells, dtype=np.int32)

    for i in range(sz):
        if nodesFTags[i] != -1:
            for j in range(len(nodesCellsId[i])):
                tags[nodesCellsId[i][j]] = nodesFTags[i]

    return tags

# Compute faces normal (np.array, shape (nb faces, 3)).
# Face normal aren't normalized.
# They are set outward from faces[2] using the vector:
#   - faces Center - cells Center
def getFacesNormal(faces, nodes, cellsCenter, meshDim):
    if meshDim == 2:
        return getFacesNormal2D(faces, nodes, cellsCenter)
    elif meshDim == 3:
        return getFacesNormal3D(faces, nodes, cellsCenter)
    else:
        raise ValueError("Invalid mesh dimension")
def getFacesNormal2D(faces, nodes, cellsCenter):
    sz = len(faces)
    normal = np.zeros((sz, 3), dtype=np.float64)
    fCenter = np.zeros(3, dtype=np.float64)
    outward = np.zeros(3, dtype=np.float64)
    tmp = np.zeros(3, dtype=np.float64)

    for i in range(sz):
        tmp[0] =  nodes[faces[i][0]][1] - nodes[faces[i][1]][1]
        tmp[1] = -nodes[faces[i][0]][0] + nodes[faces[i][1]][0]

        fCenter[:] = (nodes[faces[i][0]][:] + nodes[faces[i][1]][:]) / 2.
        outward[:] = fCenter[:] - cellsCenter[faces[i][2]][:]
        if faces[i][5] == -1:
            outward = -outward

        if (outward[0] * tmp[0] + outward[1] * tmp[1]) > 0:
            normal[i][:] =  tmp[:]
        else:
            normal[i][:] = -tmp[:]
    return normal
def getFacesNormal3D(faces, nodes, cellsCenter):
    sz = len(faces)
    u      = np.zeros(3, dtype=np.float64)
    v      = np.zeros(3, dtype=np.float64)
    fDir   = np.zeros(3, dtype=np.float64)
    normal = np.zeros((sz, 3), dtype=np.float64)

    for i in range(sz):
        n0 = faces[i][0]
        n1 = faces[i][1]
        n2 = faces[i][2]
        n3 = faces[i][3]

        # Face center
        fDir[:]  = (nodes[n0][:] + nodes[n1][:] + nodes[n2][:]) / 3

        # Vector FCenter - Left (or Right if needed)
        # Use to check that the normal is outward
        fDir[:] -= cellsCenter[faces[i][3]][:]
        if faces[i][6] == -1:
            fDir[:] = -fDir[:]

        # u and v two edge of the faces to make the cross product
        u[:] = nodes[n1][:] - nodes[n0][:]
        v[:] = nodes[n2][:] - nodes[n0][:]

        fn = np.cross(u, v) / 2
        tmp = fn[0]*fDir[0] + fn[1]*fDir[1] + fn[2]*fDir[2]
        if tmp > 0:
            normal[i][:] = fn[:]
        else:
            normal[i][:] = -fn[:]

    return normal

# Compute least square weight.
def lastSquaresWeight(nodes, nodesCellsId, nbNodesLoc, cellsCenter,
                      nbCellsLoc, faces, meshDim):
    sz = len(nodes)
    Rx  = np.zeros(sz)
    Ry  = np.zeros(sz)
    Rz  = np.zeros(sz)
    Ixx = np.zeros(sz)
    Iyy = np.zeros(sz)
    Izz = np.zeros(sz)
    Ixy = np.zeros(sz)
    Ixz = np.zeros(sz)
    Iyz = np.zeros(sz)
    nb  = np.zeros(sz, dtype=np.int8)

    alpha = [l.copy() for l in nodesCellsId]
    for i in range(len(alpha)):
        for j in range(len(alpha[i])):
            alpha[i][j] = 0.

    # if nodes id < nbNodesLoc, it is an owned nodes. Thus the surrounding
    # cells are known.
    for i in range(nbNodesLoc):
        for j in range(len(nodesCellsId[i])):
            rx = cellsCenter[nodesCellsId[i][j]][0] - nodes[i][0]
            ry = cellsCenter[nodesCellsId[i][j]][1] - nodes[i][1]
            rz = cellsCenter[nodesCellsId[i][j]][2] - nodes[i][2]
            Ixx[i] += (rx * rx)
            Iyy[i] += (ry * ry)
            Izz[i] += (rz * rz)
            Ixy[i] += (rx * ry)
            Ixz[i] += (rx * rz)
            Iyz[i] += (ry * rz)
            Rx[i] += rx
            Ry[i] += ry
            Rz[i] += rz
            nb[i] += 1

    # if nodes >= nbNodesLoc, we need to check if the node isn't on the
    # partition limits. Because we need those weight.
    # Just need to check if one cells id  is below nbCellsLoc.
    for i in range(nbNodesLoc, sz):
        isLimits = False
        for j in range(len(nodesCellsId[i])):
            if nodesCellsId[i][j] < nbCellsLoc:
                isLimits = True
        if isLimits:
            for j in range(len(nodesCellsId[i])):
                rx = cellsCenter[nodesCellsId[i][j]][0] - nodes[i][0]
                ry = cellsCenter[nodesCellsId[i][j]][1] - nodes[i][1]
                rz = cellsCenter[nodesCellsId[i][j]][2] - nodes[i][2]
                Ixx[i] += (rx * rx)
                Iyy[i] += (ry * ry)
                Izz[i] += (rz * rz)
                Ixy[i] += (rx * ry)
                Ixz[i] += (rx * rz)
                Iyz[i] += (ry * rz)
                Rx[i] += rx
                Ry[i] += ry
                Rz[i] += rz
                nb[i] += 1

    if meshDim == 2:
        # Remove nodes from border from the least square.
        for i in range(len(faces)):
            if faces[i][4] != -1:
                nb[faces[i][0]] = 0
                nb[faces[i][1]] = 0

        # Compute the weight
        for i in range(sz):
            if nb[i] != 0:
                D = Ixx[i]*Iyy[i] - Ixy[i]*Ixy[i]
                lambdaX = (Ixy[i]*Ry[i] - Iyy[i]*Rx[i]) / D
                lambdaY = (Ixy[i]*Rx[i] - Ixx[i]*Ry[i]) / D
                for j in range(len(nodesCellsId[i])):
                    rx = cellsCenter[nodesCellsId[i][j]][0] - nodes[i][0]
                    ry = cellsCenter[nodesCellsId[i][j]][1] - nodes[i][1]
                    coeff = (1 + lambdaX*rx + lambdaY*ry)\
                            / (nb[i] + lambdaX*Rx[i] + lambdaY*Ry[i])
                    alpha[i][j] = coeff
            else:
                alpha[i] = []
    elif meshDim == 3:
        # Remove nodes from border from the least square.
        for i in range(len(faces)):
            if faces[i][5] != -1:
                nb[faces[i][0]] = 0
                nb[faces[i][1]] = 0
                nb[faces[i][2]] = 0

        # Compute the weight
        for i in range(sz):
            if nb[i] != 0:
                D = Ixx[i]*Iyy[i]*Izz[i] + 2*Ixy[i]*Ixz[i]*Iyz[i] \
                        - Ixx[i]*Iyz[i]*Iyz[i] \
                        - Iyy[i]*Ixz[i]*Ixz[i] \
                        - Izz[i]*Ixy[i]*Ixy[i]

                lambdaX  = (Iyz[i]*Iyz[i] - Iyy[i]*Izz[i])*Rx[i]
                lambdaX += (Ixy[i]*Izz[i] - Ixz[i]*Iyz[i])*Ry[i]
                lambdaX += (Ixz[i]*Iyy[i] - Ixy[i]*Iyz[i])*Rz[i]
                lambdaX = lambdaX / D

                lambdaY  = (Ixy[i]*Izz[i] - Ixz[i]*Iyz[i])*Rx[i]
                lambdaY += (Ixz[i]*Ixz[i] - Ixx[i]*Izz[i])*Ry[i]
                lambdaY += (Iyz[i]*Ixx[i] - Ixz[i]*Ixy[i])*Rz[i]
                lambdaY = lambdaY / D

                lambdaZ  = (Ixz[i]*Iyy[i] - Ixy[i]*Iyz[i])*Rx[i]
                lambdaZ += (Iyz[i]*Ixx[i] - Ixz[i]*Ixy[i])*Ry[i]
                lambdaZ += (Ixy[i]*Ixy[i] - Ixx[i]*Iyy[i])*Rz[i]
                lambdaZ = lambdaZ / D

                for j in range(len(nodesCellsId[i])):
                    rx = cellsCenter[nodesCellsId[i][j]][0] - nodes[i][0]
                    ry = cellsCenter[nodesCellsId[i][j]][1] - nodes[i][1]
                    rz = cellsCenter[nodesCellsId[i][j]][2] - nodes[i][2]
                    coeff = (1 + lambdaX*rx + lambdaY*ry + lambdaZ*rz)\
                            / (nb[i] + lambdaX*Rx[i] + lambdaY*Ry[i] + lambdaZ*Rz[i])
                    alpha[i][j] = coeff
            else:
                alpha[i] = []
    else:
        raise ValueError("Invalid mesh dimension")

    return alpha

# Compute min and max h (edges) (L2 norm).
def minMaxEdgesLength(COMM, cells, nodes, nbCellsLoc):
    def nrm(ptsA, ptsB):
        x = ptsA[0] - ptsB[0]
        y = ptsA[1] - ptsB[1]
        z = ptsA[2] - ptsB[2]
        return np.sqrt(x*x + y*y + z*z)

    hma = 0
    hmi = nrm(nodes[0], nodes[1])

    for i in range(nbCellsLoc):
        nbPts = len(cells[i])
        for j in range(1, nbPts):
            for k in range(j):
                tmp = nrm(nodes[cells[i][j]], nodes[cells[i][k]])
                if tmp < hmi:
                    hmi = tmp
                if tmp > hma:
                    hma = tmp

    hmi = COMM.allreduce(hmi, op=MPI.MIN)
    hma = COMM.allreduce(hma, op=MPI.MAX)
    return hmi, hma

# Extract from faces element which:
#   - has is physical property is in rulesPhys
#   - has is cellsTags in rulesCells
def rulesSeclectFaces(faces, rulesPhys, rulesCells,
                      cellsTags, nbOwd, nbShd,  meshDim):
    sol = []
    if meshDim == 2:
        idL = 2
        idR = 3
        idP = 4
    elif meshDim == 3:
        idL = 3
        idR = 4
        idP = 5
    else:
        raise ValueError("Invalid mesh dimension")

    lst = []
    # Valid local faces
    for i in range(nbOwd):
        f  = faces[i]
        vP = f[idP]
        vL = cellsTags[f[idL]]
        vR = cellsTags[f[idR]]
        if vP in rulesPhys and (vR, vL) in rulesCells:
            lst.append(i)
    sol.append(np.array(lst))

    lst = []
    # Valid shared faces
    for i in range(nbOwd, nbOwd+nbShd):
        f  = faces[i]
        vP = f[idP]
        vL = cellsTags[f[idL]]
        vR = cellsTags[f[idR]]
        if vP in rulesPhys and (vR, vL) in rulesCells:
            lst.append(i)
    sol.append(np.array(lst))

    return sol
################################################################################
#                                           Numbering                          #
################################################################################

# Give a MPI aware local numbering from a global one.
# Return 2 array:
#   - lstLocSizes is of size SIZE
#   - lstLocSizes[RANK] is the fist id used by a given process
#   - locId[i] is the new numbering of the i-th object
def globToLocNumbering(SIZE, RANK, COMM, globId, nbLocObj):
    # Get list of all local size.
    lstLocSizes = np.zeros(SIZE+1, np.int32)
    lstLocSizes[RANK+1] = nbLocObj
    lstLocSizes = COMM.allreduce(lstLocSizes, op=MPI.SUM)
    lstLocSizes = lstLocSizes.cumsum().astype(np.int32)

    # Share in a vector the local position to sync them
    syncId = np.zeros(lstLocSizes[-1], np.int32)
    for i in range(nbLocObj):
        syncId[globId[i]] = lstLocSizes[RANK]+i
    syncId = COMM.allreduce(syncId, op=MPI.SUM)
    locId = syncId[globId]

    return lstLocSizes, locId

# Return:
#   - the number of faces owned by the MPI process.
#   - the number of shared faces owned by the current MPI process.
def facesNbOwnershipShared(SIZE, RANK, COMM, faces, cellsGlobId,
                           nbInnerFaces, nbSharedFaces):
    nbShared = 0
    nbOwned = nbInnerFaces
    for i in range(nbInnerFaces, nbInnerFaces+nbSharedFaces):
        nL = cellsGlobId[faces[i][2]]
        nR = cellsGlobId[faces[i][3]]
        if nL > nR:
            nbOwned += 1
            nbShared += 1
    return nbOwned, nbShared

# Reduce all owned size in an array.
def facesSizesArray(SIZE, RANK, COMM, nbOwned):
    ownedSizes = np.zeros(SIZE+1, dtype=np.int32)
    ownedSizes[RANK+1] = nbOwned
    ownedSizes =  COMM.allreduce(ownedSizes, op=MPI.SUM)
    ownedSizes = ownedSizes.cumsum()
    return ownedSizes

# Setup faces numbering for MPI usage.
def facesLocFacesNum(SIZE, RANK, COMM, faces, cellsGlobId, ownedSizes,
            nbFacesToSync, nbInnerFaces, nbSharedFaces):
    # Sync all sizes
    nbSyncArray = np.zeros(SIZE+1, dtype=np.int32)
    nbSyncArray[RANK+1] = nbFacesToSync
    nbSyncArray =  COMM.allreduce(nbSyncArray, op=MPI.SUM)
    nbSyncArray = nbSyncArray.cumsum()

    # Shared nodes of shared faces
    pos = 0
    facesComm = np.zeros((nbSyncArray[-1], 3), dtype=np.int32)
    for i in range(nbInnerFaces, nbInnerFaces+nbSharedFaces):
        nL = cellsGlobId[faces[i][2]]
        nR = cellsGlobId[faces[i][3]]
        if nL > nR:
            facesComm[nbSyncArray[RANK]+pos][0] = \
                    ownedSizes[RANK] + nbInnerFaces + pos
            facesComm[nbSyncArray[RANK]+pos][1] = nL
            facesComm[nbSyncArray[RANK]+pos][2] = nR
            pos += 1
    facesComm = COMM.allreduce(facesComm, op=MPI.SUM)

    # Set a new id for faces
    pos = 0
    facesCalcId = -np.ones(nbInnerFaces + nbSharedFaces, dtype=np.int32)
    for i in range(nbInnerFaces):
        facesCalcId[i] = pos + ownedSizes[RANK]
        pos += 1
    # Get shared id on shared faces
    for i in range(nbSharedFaces):
        nL = cellsGlobId[faces[i + nbInnerFaces][2]]
        nR = cellsGlobId[faces[i + nbInnerFaces][3]]
        if nL > nR:
            assert(facesCalcId[i+nbInnerFaces] == -1)
            facesCalcId[i+nbInnerFaces] = pos + ownedSizes[RANK]
            pos += 1
        else:
            for j in range(len(facesComm)):
                if facesComm[j][1] == nR and facesComm[j][2] == nL:
                    assert(facesCalcId[i+nbInnerFaces] == -1)
                    facesCalcId[i+nbInnerFaces] = facesComm[j][0]
    return facesCalcId

# Compute faces numbering to be consistent between different MPI executions.
def facesGlobFacesNum(SIZE, RANK, COMM,
            faces, nodesGlobId, cellsGlobId, ownedSizes,
            nbInnerFaces, nbSharedFaces):

    # In an array, put global nA and nB for all faces
    pos = 0
    shareAll = np.zeros((ownedSizes[SIZE], 2))
    for i in range(nbInnerFaces):
        shareAll[pos + ownedSizes[RANK]][0] = nodesGlobId[faces[i][0]]
        shareAll[pos + ownedSizes[RANK]][1] = nodesGlobId[faces[i][1]]
        pos += 1
    for i in range(nbInnerFaces, nbInnerFaces + nbSharedFaces):
        nL = cellsGlobId[faces[i][2]]
        nR = cellsGlobId[faces[i][3]]
        if nL > nR:
            shareAll[pos + ownedSizes[RANK]][0] = nodesGlobId[faces[i][0]]
            shareAll[pos + ownedSizes[RANK]][1] = nodesGlobId[faces[i][1]]
            pos += 1
    shareAll = COMM.allreduce(shareAll, op=MPI.SUM)

    # Use tuple to sort them and shared them in a new array
    sortedAll = np.zeros((ownedSizes[SIZE], 2))
    if RANK == 0:
        fNodes = []
        for i in range(ownedSizes[SIZE]):
            nA = shareAll[i][0]
            nB = shareAll[i][1]
            if nA < nB:
                fNodes.append((nA, nB))
            else:
                fNodes.append((nB, nA))
        fNodes.sort()

        for i in range(ownedSizes[SIZE]):
            sortedAll[i][0] = fNodes[i][0]
            sortedAll[i][1] = fNodes[i][1]
    sortedAll = COMM.allreduce(sortedAll, op=MPI.SUM)

    # Transfer of values form sortedAll to facesGlobId
    def findId(A, B):
        if A > B:
            A, B = B, A
        for i in range(len(sortedAll)):
            if sortedAll[i][0] == A and sortedAll[i][1] == B:
                return i
        assert(False)
    facesGlobId = -np.ones(len(faces))
    for i in range(len(faces)):
        A = nodesGlobId[faces[i][0]]
        B = nodesGlobId[faces[i][1]]
        fId = findId(A, B)
        facesGlobId[i] = fId
    return facesGlobId
################################################################################
#                                           Diamond                            #
################################################################################

# Compute diamond volume and normal.
#
# vol[i]: diamond volume of the faces i
# normals[i]: diamond normal(s) of the faces i
def getInnerDiamond(faces, nodes, cellsCenter, meshDim):
    if meshDim == 2:
        return getInnerDiamond2(faces, nodes, cellsCenter)
    elif meshDim == 3:
        return getInnerDiamond3(faces, nodes, cellsCenter)
    else:
        raise ValueError("Invalid mesh dimension")

# Orientation:
#   - X: R-L
#   - Y: B-A
# Should follow global synced orientation when needed.
#
# Each diamond have 1 normal (1 array of size 3)
def getInnerDiamond2(faces, nodes, cellsCenter):
    sz = len(faces)
    vol = np.zeros(sz, dtype=np.float64)
    normal = np.zeros((sz, 3), dtype=np.float64)
    vecs = np.zeros((2, 3), dtype=np.float64)
    tmp = np.zeros(3, dtype=np.float64)

    for i in range(sz):
        if faces[i][3] != -1:
            vecs[0] = nodes[faces[i][1]] - nodes[faces[i][0]]
            vecs[1] = cellsCenter[faces[i][3]] - cellsCenter[faces[i][2]]

            # Sync normals direction with shared faces
            if faces[i][5] == -1:
                vecs[1] = -vecs[1]

            # Check the right hand rule.
            tmpV = (vecs[1][0] * vecs[0][1] - vecs[1][1] * vecs[0][0]) / 2.
            tmp[0] = -vecs[1][1]
            tmp[1] =  vecs[1][0]
            tmp[2] = 0.0

            if tmpV < 0.0:
                vol[i] = -tmpV
                normal[i][:] = -tmp[:]
            else:
                vol[i] = tmpV
                normal[i][:] = tmp[:]
        else:
            fCenter = (nodes[faces[i][1]] + nodes[faces[i][0]]) / 2

            vecs[0] = nodes[faces[i][1]] - nodes[faces[i][0]]
            vecs[1] = 2. * fCenter - 2. * cellsCenter[faces[i][2]]

            # Sync normals direction with shared faces
            if faces[i][5] == -1:
                vecs[1] = -vecs[1]

            # Check the right hand rule.
            tmpV = (vecs[1][0] * vecs[0][1] - vecs[1][1] * vecs[0][0]) / 2.
            tmp[0] = -vecs[1][1]
            tmp[1] =  vecs[1][0]
            tmp[2] = 0.0

            if tmpV < 0.0:
                vol[i] = -tmpV
                normal[i][:] = -tmp[:]
            else:
                vol[i] = tmpV
                normal[i][:] = tmp[:]

    return vol, normal

# Each diamond have 6 normal, one for each faces
# (6 array of size 3).
#
# Diamond normal order:
#   - [A, B, L], [B, C, L], [C, A, L]
#     [A, B, R], [B, C, R], [C, A, R]
#   - Node A -> normal 1 & 4
#   - Node B -> normal 2 & 5
#   - Node C -> normal 0 & 3
def getInnerDiamond3(faces, nodes, cellsCenter):
    sz = len(faces)
    vol = np.zeros(sz, dtype=np.float64)
    normals = np.zeros((sz, 6, 3), dtype=np.float64)
    fCenter = np.zeros(3, np.float64)
    dCenter = np.zeros(3, np.float64)
    pR = np.zeros(3, np.float64)

    u = np.zeros(3, np.float64)
    v = np.zeros(3, np.float64)
    w = np.zeros(3, np.float64)

    for i in range(sz):
        f = faces[i]
        fCenter[:] = (nodes[f[0]][:] + nodes[f[1]][:] + nodes[f[2]][:]) / 3
        if faces[i][4] == -1:
            pR[:] = 2*fCenter[:] - cellsCenter[f[3]][:]
            # pR[:] = fCenter[:]
        else:
            pR[:] = cellsCenter[f[4]][:]
        dCenter[:] = (nodes[f[0]][:] + nodes[f[1]][:] + nodes[f[2]][:] + pR[:] + cellsCenter[f[3]]) / 5.

        u[:] = nodes[f[1]][:] - nodes[f[0]][:]
        v[:] = nodes[f[2]][:] - nodes[f[0]][:]
        w[:] = pR[:] - cellsCenter[f[3]][:]

        vFace = np.cross(u, v) / 2
        vol[i] = np.abs(vFace[0]*w[0] + vFace[1]*w[1] + vFace[2]*w[2]) / 3.

        # Normals
        nodesCpls = [(f[0], f[1], 0), (f[1], f[2], 1), (f[2], f[0], 2)]
        for (nA, nB, pos) in nodesCpls:
            # With Left
            u[:] = nodes[nA][:] - cellsCenter[f[3]][:]
            v[:] = nodes[nB][:] - cellsCenter[f[3]][:]
            w[:] = (nodes[nA][:] + nodes[nB][:] + cellsCenter[f[3]][:])/3 - fCenter[:]
            # w[:] = (nodes[nA][:] + nodes[nB][:] + cellsCenter[f[3]][:])/3 - dCenter[:]

            nrm = np.cross(u, v) / 2.
            sign = nrm[0]*w[0] + nrm[1]*w[1] + nrm[2]*w[2]

            if sign >= 0.:
                normals[i][pos][:] =  nrm[:]
            else:
                normals[i][pos][:] = -nrm[:]

            # With Right
            u[:] = nodes[nA][:] - pR[:]
            v[:] = nodes[nB][:] - pR[:]
            w[:] = (nodes[nA][:] + nodes[nB][:] + pR[:])/3 - fCenter[:]
            # w[:] = (nodes[nA][:] + nodes[nB][:] + pR[:])/3 - dCenter[:]

            nrm = np.cross(u, v) / 2.
            sign = nrm[0]*w[0] + nrm[1]*w[1] + nrm[2]*w[2]

            if sign >= 0.:
                normals[i][pos+3][:] =  nrm[:]
            else:
                normals[i][pos+3][:] = -nrm[:]

    return vol, normals
################################################################################
#                                           Split diamond                      #
################################################################################

# Compute volume and normal used in splitted diamond scheme
# Should follow global synced orientation when needed.
#
# vol[i][0]: left half diamond volume of the faces i
# vol[i][1]: right half diamond volume of the faces i
#
# normal[i]: normals associated with faces i
def getSplitDiamond(meshDim, faces, nodes, cellsCenter, nbDiamToCompute):
    if meshDim == 2:
            return getSplitDiamond2(faces, nodes, cellsCenter, nbDiamToCompute)
    elif meshDim == 3:
        return getSplitDiamond3(faces, nodes, cellsCenter, nbDiamToCompute)
    else:
        raise ValueError("Invalid mesh dimension")
# Orientation:
#   - X: R-F and F-L
#   - Y: B-A
#
# normals[i][0]: left half diamond normal of the faces i (array of size 3)
# normals[i][1]: right half diamond normal of the faces i (array of size 3)
def getSplitDiamond2(faces, nodes, cellsCenter, nbDiamToCompute):
    vol = np.zeros((nbDiamToCompute, 2), dtype=np.float64)
    normals = np.zeros((nbDiamToCompute, 2, 3), dtype=np.float64)

    faceCenter = np.zeros(3, dtype=np.float64)
    vecs = np.zeros((2, 3), dtype=np.float64)
    BA = np.zeros(3, dtype=np.float64)
    nrmL = np.zeros(3, dtype=np.float64)
    nrmR = np.zeros(3, dtype=np.float64)

    for i in range(nbDiamToCompute):
        if faces[i][4] == -1:
            idA = faces[i][0]
            idB = faces[i][1]
            faceCenter[:] = (nodes[idA][:] + nodes[idB][:]) / 2
            BA[:] = nodes[idB][:] - nodes[idA]

            # Sync normals direction with shared faces
            if faces[i][5] == 1:
                vecs[0][:] = faceCenter[:] - cellsCenter[faces[i][2]][:]
                vecs[1][:] = cellsCenter[faces[i][3]][:] - faceCenter[:]
            else:
                vecs[0][:] = - faceCenter[:] + cellsCenter[faces[i][2]][:]
                vecs[1][:] = - cellsCenter[faces[i][3]][:] + faceCenter[:]

            # Check the right hand rule.
            volL = (vecs[0][0] * BA[1] - vecs[0][1] * BA[0]) / 2
            volR = (vecs[1][0] * BA[1] - vecs[1][1] * BA[0]) / 2

            nrmL[0] = -vecs[0][1]
            nrmL[1] =  vecs[0][0]
            nrmR[0] = -vecs[1][1]
            nrmR[1] =  vecs[1][0]

            if volL < 0.0:
                vol[i][0] = - volL
                vol[i][1] = - volR
                normals[i][0][:] = -nrmL[:]
                normals[i][1][:] = -nrmR[:]
            else:
                vol[i][0] = volL
                vol[i][1] = volR
                normals[i][0][:] = nrmL[:]
                normals[i][1][:] = nrmR[:]

    return vol, normals
# 3 normals: nABF, nBCF and nCAF.
# They are all equal thus only one is computed and stored.
def getSplitDiamond3(faces, nodes, cellsCenter, nbDiamToCompute):
    vol = np.zeros((nbDiamToCompute, 2), dtype=np.float64)
    normals = np.zeros((nbDiamToCompute, 3), dtype=np.float64)

    pA = np.zeros(3, dtype=np.float64)
    pB = np.zeros(3, dtype=np.float64)
    pC = np.zeros(3, dtype=np.float64)
    pL = np.zeros(3, dtype=np.float64)
    pR = np.zeros(3, dtype=np.float64)
    pF = np.zeros(3, dtype=np.float64)
    nF1 = np.zeros(3, dtype=np.float64)
    nF2 = np.zeros(3, dtype=np.float64)
    nF3 = np.zeros(3, dtype=np.float64)

    for i in range(nbDiamToCompute):
        if faces[i][5] == -1:
            pA[:] = nodes[faces[i][0]][:]
            pB[:] = nodes[faces[i][1]][:]
            pC[:] = nodes[faces[i][2]][:]
            pL[:] = cellsCenter[faces[i][3]][:]
            pR[:] = cellsCenter[faces[i][4]][:]

            tmp = np.cross(pB - pA, pC - pA) / 2
            pF[:] = (pA[:] + pB[:] + pC[:]) / 3.

            # Volumes
            vL  = (tmp[0] * (pL[0] - pA[0]) \
                    + tmp[1] * (pL[1] - pA[1]) \
                    + tmp[2] * (pL[2] - pA[2])) / 3.
            if vL < 0.:
                vol[i][0] = - vL
            else:
                vol[i][0] =   vL

            vR  = (tmp[0] * (pR[0] - pA[0]) \
                    + tmp[1] * (pR[1] - pA[1]) \
                    + tmp[2] * (pR[2] - pA[2])) / 3.
            if vR < 0.:
                vol[i][1] = - vR
            else:
                vol[i][1] =   vR

            ### Normales
            nF1[:] = 0.5 * np.cross(pB - pA, pF - pA)[:]

            tmp = (nF1[0] * (pF[0] - pL[0]) \
                    + nF1[1] * (pF[1] - pL[1]) \
                    + nF1[2] * (pF[2] - pL[2]))
            if tmp < 0.:
                   normals[i][:] = - nF1[:]
            else:
                   normals[i][:] =   nF1[:]

    return vol, normals

################################################################################
#                                           Sync subset of faces id            #
################################################################################
# Get the list of Owned and Borrowed idices from shared elements of
# listOfFacesList.
# Each of those list is split bassed on the neighbouring MPI
# process.
def getFaceNeig(listOfFacesList, faces, nbInnerFaces, nbSharedFaces,
                mpiSize, mpiCellsNb, cellsMaps, meshDim):
    if meshDim == 2:
        posR  = 3
        posOw = 5
    elif meshDim == 3:
        posR  = 4
        posOw = 6
    else:
        raise ValueError("Invalid mesh dimension")

    neigLstOwnd = [[] for i in range(mpiSize)]
    neigLstBrwd = [[] for i in range(mpiSize)]
    for lstF in listOfFacesList:
        for f in lstF[1]:

            # Should never append
            if f < nbInnerFaces:
                assert(False)

            v = cellsMaps[faces[f][posR]]
            j = 0
            # Get The MPI process of the Rigth cells
            while mpiCellsNb[j] <= v:
                j += 1
            if faces[f][posOw] == 1:
                neigLstOwnd[j-1].append(f)
            else:
                neigLstBrwd[j-1].append(f)

    return neigLstOwnd, neigLstBrwd

# Get the number of inner faces, shared owned faces and borrowed faces.
def getFacesNb(listOfFacesList, neigLstOwnd, neigLstBrwd):
    nbLocalF     = 0
    nbToSyncOwnd = 0
    nbToSyncBrwd = 0

    for lstF in listOfFacesList:
        nbLocalF += len(lstF[0])
    for lstO in neigLstOwnd:
        nbToSyncOwnd += len(lstO)
    for lstB in neigLstBrwd:
        nbToSyncBrwd += len(lstB)

    return nbLocalF, nbToSyncOwnd, nbToSyncBrwd

# Synchronize the remap of the borrowed indices with those recieved.
# The results is store in facesCalcId.
def syncBrwRemap(facesMap, oldSign, rcvdSign, mpiSize):
    assert(len(oldSign) == len(rcvdSign))
    for descF in oldSign:
        j = 0
        while descF[1] != rcvdSign[j][2] and descF[2] != rcvdSign[j][1]:
            j += 1
        facesMap[descF[0]] = rcvdSign[j][0]

# Remap local faces.
def localFacesRemap(listOfFacesList, facesCalcId, nbFOwnd,
                    nbFacesArray, rank):
    vId = 0

    for lstF in listOfFacesList:
        for f in lstF[0]:
            facesCalcId[f] = nbFacesArray[rank] + vId
            vId += 1

    assert(vId == nbFOwnd)
# Remap owned faces.
# Return a list of tuple, for each shared faces:
#     (newId, glob left id, glob right id)
#     Those element are the signature of the exchange.
def ownedRemapAndSignature(neigLstOwnd, faces, facesCalcId, cellsGlobId, nbFacesArray,
                           nbFOwnd, nbFToSyncOwnd, rank, mpiSize, meshDim):
    if meshDim == 2:
        posL  = 2
        posR  = 3
    elif meshDim == 3:
        posL  = 3
        posR  = 4
    else:
        raise ValueError("Invalid mesh dimension")

    vId = nbFOwnd
    signOwnd = [[] for i in range(mpiSize)]

    for i in range(mpiSize):
        for f in neigLstOwnd[i]:
            facesCalcId[f] = nbFacesArray[rank] + vId
            vId += 1
            tmp = (facesCalcId[f], cellsGlobId[faces[f][posL]], cellsGlobId[faces[f][posR]])
            signOwnd[i].append(tmp)

    assert(vId == nbFOwnd + nbFToSyncOwnd)
    return signOwnd
# Build signature element for borowed elements.
def buildBrwdSignature(neigLstBrwd, faces, cellsGlobId, mpiSize, meshDim):
    if meshDim == 2:
        posL  = 2
        posR  = 3
    elif meshDim == 3:
        posL  = 3
        posR  = 4
    else:
        raise ValueError("Invalid mesh dimension")

    signBrwd = [[] for i in range(mpiSize)]

    for i in range(mpiSize):
        for f in neigLstBrwd[i]:
            tmp = (f, cellsGlobId[faces[f][posL]], cellsGlobId[faces[f][posR]])
            signBrwd[i].append(tmp)

    return signBrwd

# Renumber (and synchronize it) faces id for elements in listOfFacesList.
# listOfFacesList: list of list of faces id
#     id must be unique
# mpiCellsNb: the number of cells for each MPI process
# cellsMaps: should be cellsMaps1
# cellsGlobId: global celles id, should also be for a 1 mapping
def mappingFacesSubset(listOfFacesList, faces, nbInnerFaces, nbSharedFaces,
                       mpiCellsNb, cellsMaps1, cellsGlobId1,
                       SIZE, RANK, COMM, meshDim):

    # Get nb of faces to remap + sort them based on their mpi process
    neigLstOwnd, neigLstBrwd = \
            getFaceNeig(listOfFacesList, faces, nbInnerFaces, nbSharedFaces,
                                           SIZE, mpiCellsNb, cellsMaps1, meshDim)
    nbLocalF, nbFToSyncOwnd, nbFToSyncBrwd = \
            getFacesNb(listOfFacesList, neigLstOwnd, neigLstBrwd)
    nbFacesLoc = nbLocalF

    # Sync the size
    nbFacesArray = np.zeros(SIZE+1, np.int32)
    nbFacesArray[RANK+1] = nbFacesLoc + nbFToSyncOwnd
    nbFacesArray = COMM.allreduce(nbFacesArray, op=MPI.SUM)
    nbFacesArray = nbFacesArray.cumsum().astype(np.int32)

    # Build the remap (init to 1000000000)
    # When possible directly remap the values and when needed (shared ownd or
    # borwd) get the signature as well.
    facesMap1 = 1000000000 * np.ones(nbInnerFaces + nbSharedFaces, np.int32)
    localFacesRemap(listOfFacesList, facesMap1, nbLocalF,
                    nbFacesArray, RANK)
    signOwnd = ownedRemapAndSignature(neigLstOwnd, faces,
                                      facesMap1, cellsGlobId1, nbFacesArray,
                                      nbFacesLoc, nbFToSyncOwnd, RANK, SIZE, meshDim)
    signBrwd = buildBrwdSignature(neigLstBrwd, faces, cellsGlobId1, SIZE, meshDim)

    # Synchronisation round between all mpi process
    for i in range(SIZE):
        if len(signOwnd[i]) != 0 or len(signBrwd[i]) != 0:
            if i < RANK:
                if len(signOwnd[i]) != 0:
                    COMM.send(signOwnd[i], dest=i)
                if len(signBrwd[i]) != 0:
                    val = COMM.recv(source=i)
                    syncBrwRemap(facesMap1, signBrwd[i], val, SIZE)
            else:
                if len(signBrwd[i]) != 0:
                    val = COMM.recv(source=i)
                    syncBrwRemap(facesMap1, signBrwd[i], val, SIZE)
                if len(signOwnd[i]) != 0:
                    COMM.send(signOwnd[i], dest=i)

    return nbFacesLoc + nbFToSyncOwnd, nbFacesArray, facesMap1
