import numpy as np
from numba import jit, njit

import sys

################################################################################
#                                           Cells to Nodes                     #
################################################################################

# Get the matrix allowing to pass from cells to nodes.
def getCellsToNodesCoeff(row, col, coeff, nodesCellsId,
                         weight, nbNodesLoc, dim):
    size = 0
    for i in range(nbNodesLoc):
        for j in range(len(nodesCellsId[i])):
            if weight[i][j] != 0.:
                idC = nodesCellsId[i][j]
                idN = i
                for k in range(dim):
                    row[size] = dim*idN +k
                    col[size] = dim*idC +k
                    coeff[size] = weight[i][j]
                    size += 1
    return size

################################################################################
#                                           Cells faces                        #
################################################################################

# Apply fun on cells edges
def CellsEdges(row, col, coeff, idFaces, faces,
               nFaces, vCells, nodesCellsId,
               lsAlpha, fun, meshDim):
    if meshDim == 2:
        return CellsEdges2(row, col, coeff, idFaces, faces,
                           nFaces, vCells, nodesCellsId,
                           lsAlpha, fun)
    elif meshDim == 3:
        return CellsEdges3(row, col, coeff, idFaces, faces,
                           nFaces, vCells, nodesCellsId,
                           lsAlpha, fun)
    else:
        raise ValueError("Invalid mesh dimension")
def CellsEdges2(row, col, coeff, idFaces, faces,
                nFaces, vCells, nodesCellsId,
                lsAlpha, fun):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]
        vR = vCells[R]

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL,  alpha / 2)
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 2)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, alpha / 2)
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 2)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 2)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 2)

    return size
def CellsEdges3(row, col, coeff, idFaces, faces,
                nFaces, vCells, nodesCellsId,
                lsAlpha, fun):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        nF = nFaces[idF]
        vL = vCells[L]
        vR = vCells[R]

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL,  alpha / 3)
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 3)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, alpha / 3)
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 3)
        for j in range(len(lsAlpha[C])):
            idC = nodesCellsId[C][j]
            alpha = lsAlpha[C][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, alpha / 3)
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 3)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        nF = nFaces[idF]
        vL = vCells[L]

        if f[6] == -1:
            sig = -1.
        else:
            sig =  1.

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 3)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 3)
        for j in range(len(lsAlpha[C])):
            idC = nodesCellsId[C][j]
            alpha = lsAlpha[C][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 3)

    return size

# Apply fun on faces for nodes on the domain edges.
def CellsEdgesDiri(row, coeff, idFaces, faces,
                   nodes, nFaces, vCells,
                   lsAlpha, fun, meshDim):
    if meshDim == 2:
        return CellsEdgesDiri2(row, coeff, idFaces, faces,
                               nodes, nFaces, vCells,
                               lsAlpha, fun)
    elif meshDim == 3:
        return CellsEdgesDiri3(row, coeff, idFaces, faces,
                               nodes, nFaces, vCells,
                               lsAlpha, fun)
    else:
        raise ValueError("Invalid mesh dimension")
def CellsEdgesDiri2(row, coeff, idFaces, faces,
                    nodes, nFaces, vCells,
                    lsAlpha, fun):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]
        vR = vCells[R]

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A], nF, vL,  0.5)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[A], nF, vR, -0.5)
        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B], nF, vL,  0.5)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[B], nF, vR, -0.5)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A], nF, vL,  sig * 0.5)
        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B], nF, vL,  sig * 0.5)

    return size
def CellsEdgesDiri3(row, coeff, idFaces, faces,
                    nodes, nFaces, vCells,
                    lsAlpha, fun):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        nF = nFaces[idF]
        vL = vCells[L]
        vR = vCells[R]

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A], nF, vL,  1/3)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[A], nF, vR, -1/3)
        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B], nF, vL,  1/3)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[B], nF, vR, -1/3)
        if len(lsAlpha[C]) == 0:
            size += fun(row, coeff, L, size, nodes[C], nF, vL,  1/3)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[C], nF, vR, -1/3)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        nF = nFaces[idF]
        vL = vCells[L]

        if f[6] == -1:
            sig = -1.
        else:
            sig =  1.

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A], nF, vL,  sig * 1/3)
        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B], nF, vL,  sig * 1/3)
        if len(lsAlpha[C]) == 0:
            size += fun(row, coeff, L, size, nodes[C], nF, vL,  sig * 1/3)

    return size

################################################################################
#                                           Diamond scheme                     #
################################################################################

# Apply fun on diamond
def ExDiamond(meshDim, row, col, coeff, idFaces, faces,
              nFaces, nDiam, vDiam, vCells,
              nodesCellsId, lsAlpha,
              fun, funArgs):
    if meshDim == 2:
        return ExDiamond2(row, col, coeff, idFaces, faces,
                          nFaces, nDiam, vDiam, vCells,
                          nodesCellsId, lsAlpha,
                          fun, funArgs)
    elif meshDim == 3:
        return ExDiamond3(row, col, coeff, idFaces, faces,
                          nFaces, nDiam, vDiam, vCells,
                          nodesCellsId, lsAlpha,
                          fun, funArgs)
    else:
        raise ValueError("Invalid mesh dimension")
def ExDiamond2(row, col, coeff, idFaces, faces,
               nFaces, nDiam, vDiam, vCells,
               nodesCellsId, lsAlpha,
               fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., idF, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., idF, funArgs)

        size += fun(row, col, coeff, R, R, size,
                    nF, nF, vR, vD, -1., idF, funArgs)
        size += fun(row, col, coeff, R, L, size,
                    nF, nF, vR, vD,  1., idF, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD, -lsAlpha[A][j],
                        idF, funArgs)
            size += fun(row, col, coeff, R, idN, size,
                        nF, nD, vR, vD,  lsAlpha[A][j],
                        idF, funArgs)

        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD,  lsAlpha[B][j],
                        idF, funArgs)
            size += fun(row, col, coeff, R, idN, size,
                        nF, nD, vR, vD, -lsAlpha[B][j],
                        idF, funArgs)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., idF, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., idF, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD, - sig*lsAlpha[A][j],
                        idF, funArgs)

        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD,   sig*lsAlpha[B][j],
                        idF, funArgs)

    return size
def ExDiamond3(row, col, coeff, idFaces, faces,
              nFaces, nDiam, vDiam, vCells,
              nodesCellsId, lsAlpha,
              fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        nF = nFaces[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        nDA = nDiam[idF][1] + nDiam[idF][4]
        nDB = nDiam[idF][2] + nDiam[idF][5]
        nDC = nDiam[idF][0] + nDiam[idF][3]

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., idF, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., idF, funArgs)

        size += fun(row, col, coeff, R, R, size,
                    nF, nF, vR, vD, -1., idF, funArgs)
        size += fun(row, col, coeff, R, L, size,
                    nF, nF, vR, vD,  1., idF, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nDA, vL, vD, -lsAlpha[A][j],
                        idF, funArgs)
            size += fun(row, col, coeff, R, idN, size,
                        nF, nDA, vR, vD,  lsAlpha[A][j],
                        idF, funArgs)

        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nDB, vL, vD, -lsAlpha[B][j],
                        idF, funArgs)
            size += fun(row, col, coeff, R, idN, size,
                        nF, nDB, vR, vD,  lsAlpha[B][j],
                        idF, funArgs)

        for j in range(len(lsAlpha[C])):
            idN = nodesCellsId[C][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nDC, vL, vD, -lsAlpha[C][j],
                        idF, funArgs)
            size += fun(row, col, coeff, R, idN, size,
                        nF, nDC, vR, vD,  lsAlpha[C][j],
                        idF, funArgs)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        nF = nFaces[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        nDA = nDiam[idF][1] + nDiam[idF][4]
        nDB = nDiam[idF][2] + nDiam[idF][5]
        nDC = nDiam[idF][0] + nDiam[idF][3]

        if f[6] == -1:
            sig = - 1.
        else:
            sig =   1.

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., idF, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., idF, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nDA, vL, vD, - sig * lsAlpha[A][j],
                        idF, funArgs)

        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nDB, vL, vD, - sig * lsAlpha[B][j],
                        idF, funArgs)

        for j in range(len(lsAlpha[C])):
            idN = nodesCellsId[C][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nDC, vL, vD, - sig * lsAlpha[C][j],
                        idF, funArgs)

    return size

# Apply fun on diamond for nodes on the domain edges.
def ExDiamondDiriNodes(meshDim, row, coeff, idFaces, faces, nodes, cells,
                       nFaces, nDiam, vDiam, vCells,
                       lsAlpha, fun, funArgs):
    if meshDim == 2:
        return ExDiamondDiriNodes2(row, coeff, idFaces, faces, nodes, cells,
                                   nFaces, nDiam, vDiam, vCells,
                                   lsAlpha, fun, funArgs)
    elif meshDim == 3:
        return ExDiamondDiriNodes3(row, coeff, idFaces, faces, nodes, cells,
                                   nFaces, nDiam, vDiam, vCells,
                                   lsAlpha, fun, funArgs)
    else:
        raise ValueError("Invalid mesh dimension")
def ExDiamondDiriNodes2(row, coeff, idFaces, faces, nodes, cells,
                        nFaces, nDiam, vDiam, vCells,
                        lsAlpha, fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A],
                        nF, nD, vL, vD, -1., funArgs)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[A],
                            nF, nD, vR, vD,  1, funArgs)

        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B],
                        nF, nD, vL, vD,  1., funArgs)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[B],
                            nF, nD, vR, vD, -1., funArgs)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.
        assert(R != -1)

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A],
                        nF, nD, vL, vD, -sig, funArgs)

        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B],
                        nF, nD, vL, vD,  sig, funArgs)

    return size
def ExDiamondDiriNodes3(row, coeff, idFaces, faces, nodes, cells,
                   nFaces, nDiam, vDiam, vCells,
                   lsAlpha, fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        nF = nFaces[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        nDA = nDiam[idF][1] + nDiam[idF][4]
        nDB = nDiam[idF][2] + nDiam[idF][5]
        nDC = nDiam[idF][0] + nDiam[idF][3]

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A],
                        nF, nDA, vL, vD, -1, funArgs)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[A],
                            nF, nDA, vR, vD,  1, funArgs)
        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B],
                        nF, nDB, vL, vD, -1, funArgs)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[B],
                            nF, nDB, vR, vD,  1, funArgs)
        if len(lsAlpha[C]) == 0:
            size += fun(row, coeff, L, size, nodes[C],
                        nF, nDC, vL, vD, -1, funArgs)
            if R != -1:
                size += fun(row, coeff, R, size, nodes[C],
                            nF, nDC, vR, vD,  1, funArgs)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        nF = nFaces[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        nDA = nDiam[idF][1] + nDiam[idF][4]
        nDB = nDiam[idF][2] + nDiam[idF][5]
        nDC = nDiam[idF][0] + nDiam[idF][3]

        if f[6] == -1:
            sig = -1.
        else:
            sig =  1.

        if len(lsAlpha[A]) == 0:
            size += fun(row, coeff, L, size, nodes[A],
                        nF, nDA, vL, vD, -sig, funArgs)
        if len(lsAlpha[B]) == 0:
            size += fun(row, coeff, L, size, nodes[B],
                        nF, nDB, vL, vD, -sig, funArgs)
        if len(lsAlpha[C]) == 0:
            size += fun(row, coeff, L, size, nodes[C],
                        nF, nDC, vL, vD, -sig, funArgs)

    return size

# Apply dirichlet diamond conditions fun on the edges for the matrix
# operator.
def ExDiamondDiriFaces(meshDim, row, col, coeff, idFaces, faces,
                       nFaces, vDiam, vCells,
                       fun, funArgs):
    if meshDim == 2:
        return ExDiamondDiriFaces2(row, col, coeff, idFaces, faces,
                                   nFaces, vDiam, vCells,
                                   fun, funArgs)
    elif meshDim == 3:
        return ExDiamondDiriFaces3(row, col, coeff, idFaces, faces,
                                   nFaces, vDiam, vCells,
                                   fun, funArgs)
    else:
        raise ValueError("Invalid mesh dimension")
def ExDiamondDiriFaces2(row, col, coeff, idFaces, faces,
                        nFaces, vDiam, vCells,
                        fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        L = f[2]
        nF = nFaces[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., idF, funArgs)

    # There is no shared face on the domain border.

    return size
def ExDiamondDiriFaces3(row, col, coeff, idFaces, faces,
                        nFaces, vDiam, vCells,
                        fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        L = f[3]
        nF = nFaces[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., idF, funArgs)

    # There is no shared face on the domain border.

    return size

# Apply dirichlet diamond conditions fun on the edges for the rhs term.
def ExDiamondDiriRhs(meshDim, row, coeff, idFaces, faces, nodes, cells,
                     nFaces, nDiam, vDiam, vCells,
                     fun, funArgs):
    if meshDim == 2:
        return ExDiamondDiriRhs2(row, coeff, idFaces, faces, nodes, cells,
                                 nFaces, nDiam, vDiam, vCells,
                                 fun, funArgs)
    elif meshDim == 3:
        return ExDiamondDiriRhs3(row, coeff, idFaces, faces, nodes, cells,
                                 nFaces, nDiam, vDiam, vCells,
                                 fun, funArgs)
    else:
        raise ValueError("Invalid mesh dimension")
def ExDiamondDiriRhs2(row, coeff, idFaces, faces, nodes, cells,
                      nFaces, nDiam, vDiam, vCells,
                      fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f  = faces[idF]
        L  = f[2]
        pA = nodes[f[0]]
        pB = nodes[f[1]]
        pL = cells[L]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        size += fun(row, coeff, L, size,
                    pL, pA, pB, nF, nD, vL, vD,
                    funArgs)

    # There is no shared face on the domain border.

    return size
def ExDiamondDiriRhs3(row, coeff, idFaces, faces, nodes, cells,
                      nFaces, nDiam, vDiam, vCells,
                      fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f  = faces[idF]
        L  = f[3]
        pA = nodes[f[0]]
        pB = nodes[f[1]]
        pC = nodes[f[2]]
        pL = cells[L]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        size += fun(row, coeff, L, size,
                    pL, pA, pB, pC, nF, nD, vL, vD,
                    funArgs)

    # There is no shared face on the domain border.

    return size

# Apply neumann diamond conditions fun on the edges for the rhs term.
def ExDiamondNeum(meshDim, row, coeff, idFaces, faces, nodes,
                  nFaces, vCells, fun, funArgs):
    if meshDim == 2:

        return ExDiamondNeum2(row, coeff, idFaces, faces, nodes,
                              nFaces, vCells, fun, funArgs)
    elif meshDim == 3:
        return ExDiamondNeum3(row, coeff, idFaces, faces, nodes,
                              nFaces, vCells, fun, funArgs)
    else:
        raise ValueError("Invalid mesh dimension")
def ExDiamondNeum2(row, coeff, idFaces, faces, nodes,
                   nFaces, vCells, fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f  = faces[idF]
        L  = f[2]
        pA = nodes[f[0]]
        pB = nodes[f[1]]
        nF = nFaces[idF]
        vL = vCells[L]

        size += fun(row, coeff, L, size,
                    pA, pB, nF, vL, funArgs)

    # There is no shared face on the domain border.

    return size
def ExDiamondNeum3(row, coeff, idFaces, faces, nodes,
                   nFaces, vCells, fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f  = faces[idF]
        L  = f[3]
        pA = nodes[f[0]]
        pB = nodes[f[1]]
        pC = nodes[f[2]]
        nF = nFaces[idF]
        vL = vCells[L]

        size += fun(row, coeff, L, size,
                    pA, pB, pC, nF, vL, funArgs)

    # There is no shared face on the domain border.

    return size
################################################################################
#                                           Split Diamond scheme               #
################################################################################

# Apply fun on diamond
# 4 matrix will be build tagged with [[1, 2], [3, 4]].
def ExSplitLR(meshDim, row, col, coeff,
              tag, idFaces, faces,
              nFaces, nDiam, nSptDiam,
              vDiam, vSptDiam, vCells,
              nodesCellsId, lsAlpha,
              fun, funArgs):
    if meshDim == 2:
        return ExSplitLR2(row, col, coeff, tag, idFaces, faces,
                          nFaces, nDiam, nSptDiam,
                          vDiam, vSptDiam, vCells,
                          nodesCellsId, lsAlpha,
                          fun, funArgs)
    elif meshDim == 3:
        return ExSplitLR3(row, col, coeff, tag, idFaces, faces,
                          nFaces, nDiam, nSptDiam,
                          vDiam, vSptDiam, vCells,
                          nodesCellsId, lsAlpha,
                          fun, funArgs)
    else:
        raise ValueError("Invalid mesh dimension")
def ExSplitLR2(row, col, coeff, tag, idFaces, faces,
               nFaces, nDiam, nSptDiam,
               vDiam, vSptDiam, vCells,
               nodesCellsId, lsAlpha,
               fun, funArgs):
    size = 0

    # Owned Left
    for idF in idFaces[0][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        F = idF

        nF = nFaces[idF]
        nD = nDiam[idF]
        nDL = nSptDiam[idF][0]
        nDR = nSptDiam[idF][1]

        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]
        vDL = vSptDiam[idF][0]
        vDR = vSptDiam[idF][1]

        if tag == 2:
            size += fun(row, col, coeff, L, F, size,
                        nF, nF, vL, vDL,  1.,
                        L, L, funArgs)
        if tag == 4:
            size += fun(row, col, coeff, F, F, size,
                        nF, nF, 1., vDL,  1.,
                        L, L, funArgs)
        if tag == 1:
            size += fun(row, col, coeff, L, L, size,
                        nF, nF, vL, vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nDL, vL, vDL, -lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nDL, vL, vDL,  lsAlpha[B][j],
                            L, L, funArgs)
        if tag == 3:
            size += fun(row, col, coeff, F, L, size,
                        nF, nF, 1., vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nDL, 1., vDL, -lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nDL, 1., vDL,  lsAlpha[B][j],
                            L, L, funArgs)

    # Shared Left
    for idF in idFaces[0][1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        F = idF

        nF = nFaces[idF]
        nD = nDiam[idF]
        nDL = nSptDiam[idF][0]

        vD = vDiam[idF]
        vL = vCells[L]
        vDL = vSptDiam[idF][0]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        if tag == 2:
            size += fun(row, col, coeff, L, F, size,
                        nF, nF, vL, vDL,  1.,
                        L, L, funArgs)
        if tag == 4:
            size += fun(row, col, coeff, F, F, size,
                        nF, nF, 1., vDL, 1.,
                        L, L, funArgs)
        if tag == 1:
            size += fun(row, col, coeff, L, L, size,
                        nF, nF, vL, vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nDL, vL, vDL, - sig * lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nDL, vL, vDL,  sig * lsAlpha[B][j],
                            L, L, funArgs)
        if tag == 3:
            size += fun(row, col, coeff, F, L, size,
                        nF, nF, 1., vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nDL, 1., vDL, -sig * lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nDL, 1., vDL,  sig * lsAlpha[B][j],
                            L, L, funArgs)

    # Owned Right
    for idF in idFaces[1][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        F = idF

        nF = nFaces[idF]
        nD = nDiam[idF]
        nDL = nSptDiam[idF][0]
        nDR = nSptDiam[idF][1]

        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]
        vDL = vSptDiam[idF][0]
        vDR = vSptDiam[idF][1]

        if tag == 2:
            size += fun(row, col, coeff, R, F, size,
                        nF, nF, vR, vDR,  1.,
                        R, R, funArgs)
        if tag == 4:
            size += fun(row, col, coeff, F, F, size,
                        nF, nF, 1., vDR,  1.,
                        R, R, funArgs)
        if tag == 1:
            size += fun(row, col, coeff, R, R, size,
                        nF, nF, vR, vDR, -1.,
                        R, R, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, R, idN, size,
                            nF, nDR, vR, vDR,  lsAlpha[A][j],
                            R, R, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, R, idN, size,
                            nF, nDR, vR, vDR, -lsAlpha[B][j],
                            R, R, funArgs)
        if tag == 3:
            size += fun(row, col, coeff, F, R, size,
                        nF, nF, 1., vDR, -1.,
                        R, R, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nDR, 1., vDR,  lsAlpha[A][j],
                            R, R, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nDR, 1., vDR, -lsAlpha[B][j],
                            R, R, funArgs)

    # No shared Right
    return size
def ExSplitLR3(row, col, coeff, tag, idFaces, faces,
               nFaces, nDiam, nSptDiam,
               vDiam, vSptDiam, vCells,
               nodesCellsId, lsAlpha,
               fun, funArgs):
    size = 0

    # Owned Left
    for idF in idFaces[0][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        F =  idF

        vD  = vDiam[idF]
        vL  = vCells[L]
        vR  = vCells[R]
        vDL = vSptDiam[idF][0]
        vDR = vSptDiam[idF][1]

        nF  = nFaces[idF]
        nDS = nSptDiam[idF]
        nLA = nDiam[idF][1] + nDS
        nLB = nDiam[idF][2] + nDS
        nLC = nDiam[idF][0] + nDS

        if tag == 2:
            size += fun(row, col, coeff, L, F, size,
                        nF, nF, vL, vDL,  1.,
                        L, L, funArgs)
        if tag == 4:
            size += fun(row, col, coeff, F, F, size,
                        nF, nF, 1., vDL,  1.,
                        L, L, funArgs)
        if tag == 1:
            size += fun(row, col, coeff, L, L, size,
                        nF, nF, vL, vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nLA, vL, vDL, -lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nLB, vL, vDL, -lsAlpha[B][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[C])):
                idN = nodesCellsId[C][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nLC, vL, vDL, -lsAlpha[C][j],
                            L, L, funArgs)
        if tag == 3:
            size += fun(row, col, coeff, F, L, size,
                        nF, nF, 1., vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nLA, vL, vDL, -lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nLB, vL, vDL, -lsAlpha[B][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[C])):
                idN = nodesCellsId[C][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nLC, vL, vDL, -lsAlpha[C][j],
                            L, L, funArgs)

    # Shared Left
    for idF in idFaces[0][1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        F =  idF

        vD  = vDiam[idF]
        vL  = vCells[L]
        vR  = vCells[R]
        vDL = vSptDiam[idF][0]
        vDR = vSptDiam[idF][1]

        nF  = nFaces[idF]
        nDS = nSptDiam[idF]
        nLA = nDiam[idF][1] + nDS
        nLB = nDiam[idF][2] + nDS
        nLC = nDiam[idF][0] + nDS

        if f[6] == -1:
            sig = -1.
        else:
            sig =  1.

        if tag == 2:
            size += fun(row, col, coeff, L, F, size,
                        nF, nF, vL, vDL,  1.,
                        L, L, funArgs)
        if tag == 4:
            size += fun(row, col, coeff, F, F, size,
                        nF, nF, 1., vDL,  1.,
                        L, L, funArgs)
        if tag == 1:
            size += fun(row, col, coeff, L, L, size,
                        nF, nF, vL, vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nLA, vL, vDL, - sig * lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nLB, vL, vDL, - sig * lsAlpha[B][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[C])):
                idN = nodesCellsId[C][j]
                size += fun(row, col, coeff, L, idN, size,
                            nF, nLC, vL, vDL, - sig * lsAlpha[C][j],
                            L, L, funArgs)
        if tag == 3:
            size += fun(row, col, coeff, F, L, size,
                        nF, nF, 1., vDL, -1.,
                        L, L, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nLA, vL, vDL, - sig * lsAlpha[A][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nLB, vL, vDL, - sig * lsAlpha[B][j],
                            L, L, funArgs)
            for j in range(len(lsAlpha[C])):
                idN = nodesCellsId[C][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nLC, vL, vDL, - sig * lsAlpha[C][j],
                            L, L, funArgs)

    # Owned Right
    for idF in idFaces[1][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        C = f[2]
        L = f[3]
        R = f[4]
        F =  idF

        vD  = vDiam[idF]
        vL  = vCells[L]
        vR  = vCells[R]
        vDL = vSptDiam[idF][0]
        vDR = vSptDiam[idF][1]

        nF  = nFaces[idF]
        nDS = nSptDiam[idF]
        nRA = nDiam[idF][4] - nDS
        nRB = nDiam[idF][5] - nDS
        nRC = nDiam[idF][3] - nDS

        if tag == 2:
            size += fun(row, col, coeff, R, F, size,
                        nF, nF, vR, vDR,  1.,
                        R, R, funArgs)
        if tag == 4:
            size += fun(row, col, coeff, F, F, size,
                        nF, nF, 1., vDR,  1.,
                        R, R, funArgs)
        if tag == 1:
            size += fun(row, col, coeff, R, R, size,
                        nF, nF, vR, vDR, -1.,
                        R, R, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, R, idN, size,
                            nF, nRA, vR, vDR,  lsAlpha[A][j],
                            R, R, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, R, idN, size,
                            nF, nRB, vR, vDR,  lsAlpha[B][j],
                            R, R, funArgs)
            for j in range(len(lsAlpha[C])):
                idN = nodesCellsId[C][j]
                size += fun(row, col, coeff, R, idN, size,
                            nF, nRC, vR, vDR,  lsAlpha[C][j],
                            R, R, funArgs)
        if tag == 3:
            size += fun(row, col, coeff, F, R, size,
                        nF, nF, 1., vDR, -1.,
                        R, R, funArgs)
            for j in range(len(lsAlpha[A])):
                idN = nodesCellsId[A][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nRA, vR, vDR,  lsAlpha[A][j],
                            R, R, funArgs)
            for j in range(len(lsAlpha[B])):
                idN = nodesCellsId[B][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nRB, vR, vDR,  lsAlpha[B][j],
                            R, R, funArgs)
            for j in range(len(lsAlpha[C])):
                idN = nodesCellsId[C][j]
                size += fun(row, col, coeff, F, idN, size,
                            nF, nRC, vR, vDR,  lsAlpha[C][j],
                            R, R, funArgs)

    # No shared Right
    return size
################################################################################
#                                           Generics matrix                    #
################################################################################

# SubFunction to write 1. * vol on the diagonal.
@njit
def cellsId3(row, col, coeff, size, pos, vol):
    pos = 3*pos
    for l in range(3):
        row[size] = pos+l
        col[size] = pos+l
        coeff[size] = vol
        size += 1
    return 3
@njit
def cellsId1(row, col, coeff, size, pos, vol):
    row[size]   = pos
    col[size]   = pos
    coeff[size] = vol
    return 1

# SubFunction to write 1. on the diagonal.
@njit
def cellsOneOnDiag(row, col, coeff, size, pos, vol):
    pos = 3*pos
    for l in range(3):
        row[size] = pos+l
        col[size] = pos+l
        coeff[size] = 1.
        size += 1
    return 3

# SubFunction to compute the curl.
# 3 by 3 matrix for curl coeff
#         0 -a3  a2
# a x =  a3   0 -a1
#       -a2  a1   0
@njit
def edgesCurl(row, col, coeff, rowId, colId, size, n, v, alpha):
    row[size] = 3*rowId
    col[size] = 3*colId +1
    coeff[size] = -alpha * n[2] / v
    size += 1
    row[size] = 3*rowId
    col[size] = 3*colId +2
    coeff[size] =  alpha * n[1] / v
    size += 1

    row[size] = 3*rowId +1
    col[size] = 3*colId
    coeff[size] =  alpha * n[2] / v
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId +2
    coeff[size] = -alpha * n[0] / v
    size += 1

    row[size] = 3*rowId +2
    col[size] = 3*colId
    coeff[size] = -alpha * n[1] / v
    size += 1
    row[size] = 3*rowId +2
    col[size] = 3*colId +1
    coeff[size] =  alpha * n[0] / v
    size += 1

    return 6
@njit
def edgesCurlZ(row, col, coeff, rowId, colId, size, n, v, alpha):
    row[size] = 3*rowId
    col[size] =   colId
    coeff[size] = alpha * n[1] / v
    size += 1

    row[size] = 3*rowId +1
    col[size] =   colId
    coeff[size] = -alpha * n[0] / v
    size += 1

    return 2

### Curl matrix
#     a1
# a = a2
#     a3
#         0 -a3  a2                      0   0  a2
# a x =  a3   0 -a1   donc en 2D on a:   0   0 -a1
#       -a2  a1   0                    -a2  a1   0
#                        -a2b2  a2b1          0
# (a x b) x en 2D donne:  a1b2 -a1b1          0
#                            0     0 -a2b2-a1b1
# 3 by 3 matrix for curl curl coeff in 2D (z = 0)
@njit
def curlCurl(row, col, coeff, rowId, colId, size,
             nA, nB, vC, vD, alpha,
             idF, args):

    w = args[idF]

    row[size] = 3*rowId
    col[size] = 3*colId
    coeff[size] = -alpha * w * nA[1] * nB[1]
    size += 1
    row[size] = 3*rowId
    col[size] = 3*colId +1
    coeff[size] =  alpha * w * nA[1] * nB[0]
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId
    coeff[size] =  alpha * w * nA[0] * nB[1]
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId +1
    coeff[size] = -alpha * w * nA[0] * nB[0]
    size += 1
    row[size] = 3*rowId +2
    col[size] = 3*colId +2
    coeff[size] = -alpha * w * (nA[0] * nB[0] + nA[1] * nB[1])
    size += 1

    return 5
def curlCurlEps(row, col, coeff, rowId, colId, size,
                   nA, nB, vC, vD, alpha,
                   useless1, useless2, useless3):

    minFact = np.min(np.abs([
        nA[1] * nB[1],
        nA[1] * nB[0],
        nA[0] * nB[1],
        nA[0] * nB[0],
        nA[0] * nB[0] + nA[1] * nB[1]
        ]))
    eps = 1e-5 * minFact

    v = 2 * vD
    row[size] = 3*rowId
    col[size] = 3*colId
    coeff[size] = - alpha * (nA[1] * nB[1] + eps) / v
    size += 1
    row[size] = 3*rowId
    col[size] = 3*colId +1
    coeff[size] =   alpha * nA[1] * nB[0] / v
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId
    coeff[size] =   alpha * nA[0] * nB[1] / v
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId +1
    coeff[size] = - alpha * (nA[0] * nB[0] + eps) / v
    size += 1
    row[size] = 3*rowId +2
    col[size] = 3*colId +2
    coeff[size] = - alpha * (nA[0] * nB[0] + nA[1] * nB[1] + eps) / v
    size += 1

    return 5
@njit
def curlCurl3(row, col, coeff, rowId, colId, size,
              nA, nB, vC, vD, alpha,
             idF, args):

    w = args[idF]

    # Diag
    row[size] = 3*rowId
    col[size] = 3*colId
    coeff[size] = -alpha * w * (nA[1] * nB[1] + nA[2] * nB[2])
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId +1
    coeff[size] = -alpha * w * (nA[0] * nB[0] + nA[2] * nB[2])
    size += 1
    row[size] = 3*rowId +2
    col[size] = 3*colId +2
    coeff[size] = -alpha * w * (nA[0] * nB[0] + nA[1] * nB[1])
    size += 1

    # Other term
    row[size] = 3*rowId
    col[size] = 3*colId +1
    coeff[size] =  alpha * w * nA[1] * nB[0]
    size += 1
    row[size] = 3*rowId
    col[size] = 3*colId +2
    coeff[size] =  alpha * w * nA[2] * nB[0]
    size += 1

    row[size] = 3*rowId +1
    col[size] = 3*colId
    coeff[size] =  alpha * w * nA[0] * nB[1]
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId +2
    coeff[size] =  alpha * w * nA[2] * nB[1]
    size += 1

    row[size] = 3*rowId +2
    col[size] = 3*colId
    coeff[size] =  alpha * w * nA[0] * nB[2]
    size += 1
    row[size] = 3*rowId +2
    col[size] = 3*colId +1
    coeff[size] =  alpha * w * nA[1] * nB[2]
    size += 1

    return 9
@njit
def curl3(row, col, coeff, rowId, colId, size,
              nA, nB, vC, vD, alpha,
              useless1, useless2, useless3):

    v = 3 * vD

    # Other term
    row[size]    = 3*rowId
    col[size]    = 3*colId +1
    coeff[size]  = -alpha * nB[2] / v
    size += 1
    row[size]    = 3*rowId
    col[size]    = 3*colId +2
    coeff[size]  =  alpha * nB[1] / v
    size += 1

    row[size]    = 3*rowId +1
    col[size]    = 3*colId
    coeff[size]  =  alpha * nB[2] / v
    size += 1
    row[size]    = 3*rowId +1
    col[size]    = 3*colId +2
    coeff[size]  = -alpha * nB[0] / v
    size += 1

    row[size]    = 3*rowId +2
    col[size]    = 3*colId
    coeff[size]  = -alpha * nB[1] / v
    size += 1
    row[size]    = 3*rowId +2
    col[size]    = 3*colId +1
    coeff[size]  =  alpha * nB[0] / v
    size += 1

    return 6
@njit
def lapl2(row, col, coeff, rowId, colId, size,
          nA, nB, vC, vD, alpha,
          idF, args):

    w = args[idF]

    # Diag
    row[size] = rowId
    col[size] = colId
    coeff[size] = - alpha * w * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2])
    size += 1

    return 1
@njit
def lapl3(row, col, coeff, rowId, colId, size,
          nA, nB, vC, vD, alpha,
          idF, args):

    w = args[idF]

    # Diag
    row[size] = rowId
    col[size] = colId
    coeff[size] = - alpha * w * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2])
    size += 1

    return 1
@njit
def laplRho2(row, col, coeff, rowId, colId, size,
             nA, nB, vC, vD, alpha,
             useless1, useless2, useless3):

    rhoAir = (useless3[useless1] + useless3[useless2])/2
    v = 2 * vD

    # Diag
    row[size] = rowId
    col[size] = colId
    coeff[size] = - rhoAir * alpha * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2]) / v
    size += 1

    return 1
@njit
def laplRho3(row, col, coeff, rowId, colId, size,
              nA, nB, vC, vD, alpha,
              useless1, useless2, useless3):

    rhoAir = (useless3[useless1] + useless3[useless2])/2
    v = 3 * vD

    # Diag
    row[size] = rowId
    col[size] = colId
    coeff[size] = - rhoAir * alpha * (nA[0] * nB[0] + nA[1] * nB[1] + nA[2] * nB[2]) / v
    size += 1

    return 1

# 3 by 3 matrix for curl curl coeff in 2D (z = 0)
@njit
def curlRhoCurl(row, col, coeff, rowId, colId, size,
              nA, nB, vC, vD, alpha,
              useless1, useless2, useless3):

    rhoAir = (useless3[3*useless1] + useless3[3*useless2])/2

    v = 2 * vD
    row[size] = 3*rowId
    col[size] = 3*colId
    coeff[size] = -rhoAir * alpha * nA[1] * nB[1] / v
    size += 1
    row[size] = 3*rowId
    col[size] = 3*colId +1
    coeff[size] =  rhoAir * alpha * nA[1] * nB[0] / v
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId
    coeff[size] =  rhoAir * alpha * nA[0] * nB[1] / v
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId +1
    coeff[size] = -rhoAir * alpha * nA[0] * nB[0] / v
    size += 1
    row[size] = 3*rowId +2
    col[size] = 3*colId +2
    coeff[size] = -rhoAir * alpha * (nA[0] * nB[0] + nA[1] * nB[1]) / v
    size += 1

    return 5

# 3 by 3 matrix for curl curl coeff in 2D (z = 0)
# Cant @njit because of the min list
def curlRhoCurlEps(row, col, coeff, rowId, colId, size,
                   nA, nB, vC, vD, alpha,
                   useless1, useless2, useless3):

    rhoAir = (useless3[3*useless1] + useless3[3*useless2])/2
    minFact = np.min(np.abs([
        nA[1] * nB[1],
        nA[1] * nB[0],
        nA[0] * nB[1],
        nA[0] * nB[0],
        nA[0] * nB[0] + nA[1] * nB[1]
        ]))
    eps = 1e-5 * minFact

    v = 2 * vD
    row[size] = 3*rowId
    col[size] = 3*colId
    coeff[size] = -rhoAir * alpha * (nA[1] * nB[1] + eps) / v
    size += 1
    row[size] = 3*rowId
    col[size] = 3*colId +1
    coeff[size] =  rhoAir * alpha * nA[1] * nB[0] / v
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId
    coeff[size] =  rhoAir * alpha * nA[0] * nB[1] / v
    size += 1
    row[size] = 3*rowId +1
    col[size] = 3*colId +1
    coeff[size] = -rhoAir * alpha * (nA[0] * nB[0] + eps) / v
    size += 1
    row[size] = 3*rowId +2
    col[size] = 3*colId +2
    coeff[size] = -rhoAir * alpha * (nA[0] * nB[0] + nA[1] * nB[1] + eps) / v
    size += 1

    return 5

# Stripped version of curlRhoCurl
@njit
def curlRhoCurlStrip(row, col, coeff, rowId, colId, size,
                     nA, nB, vC, vD, alpha,
                     p1, p2, wght):

    w = (wght[3*p1] + wght[3*p2])/2
    v = 2 * vD

    row[size] = 3*rowId +2
    col[size] = 3*colId +2
    coeff[size] = -w * alpha * (nA[0] * nB[0] + nA[1] * nB[1]) / v
    size += 1

    row[size] = 3*rowId
    col[size] = 3*colId
    coeff[size] = 1.
    size += 1

    row[size] = 3*rowId +1
    col[size] = 3*colId +1
    coeff[size] = 1.
    size += 1
    return 3

# Apply fun on cells
def ExCells(row, col, coeff, idCells, vCells, fun):
    size = 0
    for i in idCells:
        vC = vCells[i]
        size += fun(row, col, coeff, size, i, vC)
    return size

# Apply fun on cells edges
def ExCellsEdges(row, col, coeff, idFaces, faces,
              nFaces, vCells, nodesCellsId, lsAlpha,
              fun):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]
        vR = vCells[R]

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL,  alpha / 2)
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 2)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, alpha / 2)
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 2)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 2)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 2)

    return size

# Apply fun on cells edges
def ExCellsEdgesLR(row, col, coeff, idFaces, faces,
              nFaces, vCells, nodesCellsId, lsAlpha,
              fun):
    size = 0

    # Owned Left
    for idF in idFaces[0][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]
        vR = vCells[R]

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL,  alpha / 2)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, alpha / 2)
    # Shared Left
    for idF in idFaces[0][1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 2)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        L, idC, size, nF, vL, sig*alpha / 2)

    # Owned Right
    for idF in idFaces[1][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        vL = vCells[L]
        vR = vCells[R]

        for j in range(len(lsAlpha[A])):
            idC = nodesCellsId[A][j]
            alpha = lsAlpha[A][j]
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 2)
        for j in range(len(lsAlpha[B])):
            idC = nodesCellsId[B][j]
            alpha = lsAlpha[B][j]
            size += fun(row, col, coeff,
                        R, idC, size, nF, vR, -alpha / 2)

    # No shared Right
    return size

# Apply fun on diamond
def ExDiamondOld(row, col, coeff, idFaces, faces,
              nFaces, nDiam, vDiam, vCells,
              nodesCellsId, lsAlpha,
              fun, funArgs):
    size = 0

    # Faces owned
    for idF in idFaces[0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., L, R, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., L, R, funArgs)

        size += fun(row, col, coeff, R, R, size,
                    nF, nF, vR, vD, -1., L, R, funArgs)
        size += fun(row, col, coeff, R, L, size,
                    nF, nF, vR, vD,  1., L, R, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD, -lsAlpha[A][j],
                        L, R, funArgs)
            size += fun(row, col, coeff, R, idN, size,
                        nF, nD, vR, vD,  lsAlpha[A][j],
                        L, R, funArgs)

        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD,  lsAlpha[B][j],
                        L, R, funArgs)
            size += fun(row, col, coeff, R, idN, size,
                        nF, nD, vR, vD, -lsAlpha[B][j],
                        L, R, funArgs)

    # Faces shared
    for idF in idFaces[1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., L, R, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., L, R, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD, - sig*lsAlpha[A][j],
                        L, R, funArgs)

        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD,   sig*lsAlpha[B][j],
                        L, R, funArgs)

    return size

# Apply fun on diamond
def ExDiamondLR(row, col, coeff, idFaces, faces,
              nFaces, nDiam, vDiam, vCells,
              nodesCellsId, lsAlpha,
              fun, funArgs):
    size = 0

    # Owned Left
    for idF in idFaces[0][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., L, R, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., L, R, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD, -lsAlpha[A][j],
                        L, R, funArgs)
        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD,  lsAlpha[B][j],
                        L, R, funArgs)

    # Shared Left
    for idF in idFaces[0][1]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        if f[5] == -1:
            sig = -1.
        else:
            sig =  1.

        size += fun(row, col, coeff, L, R, size,
                    nF, nF, vL, vD,  1., L, R, funArgs)
        size += fun(row, col, coeff, L, L, size,
                    nF, nF, vL, vD, -1., L, R, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD, - sig*lsAlpha[A][j],
                        L, R, funArgs)
        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, L, idN, size,
                        nF, nD, vL, vD,   sig*lsAlpha[B][j],
                        L, R, funArgs)

    # Owned Right
    for idF in idFaces[1][0]:
        f = faces[idF]
        A = f[0]
        B = f[1]
        L = f[2]
        R = f[3]
        nF = nFaces[idF]
        nD = nDiam[idF]
        vD = vDiam[idF]
        vL = vCells[L]
        vR = vCells[R]

        size += fun(row, col, coeff, R, R, size,
                    nF, nF, vR, vD, -1., L, R, funArgs)
        size += fun(row, col, coeff, R, L, size,
                    nF, nF, vR, vD,  1., L, R, funArgs)

        for j in range(len(lsAlpha[A])):
            idN = nodesCellsId[A][j]
            size += fun(row, col, coeff, R, idN, size,
                        nF, nD, vR, vD,  lsAlpha[A][j],
                        L, R, funArgs)
        for j in range(len(lsAlpha[B])):
            idN = nodesCellsId[B][j]
            size += fun(row, col, coeff, R, idN, size,
                        nF, nD, vR, vD, -lsAlpha[B][j],
                        L, R, funArgs)

    # No shared Right
    return size

@njit
def supLinZ(row, col, coeff, rowId, colId, size,
              nA, nB, vC, vD, alpha,
              idF, args):

    v   = args[idF][0]
    JA0 = args[idF][1]
    JA1 = args[idF][2]
    JA2 = args[idF][3]

    mu0 = 4 * np.pi * 1e-7
    Ec  = 1e-4
    Jc  =  1e7
    n   =   10

    curlA = np.array([[    0., -nA[2],  nA[1]],
                      [ nA[2],     0., -nA[0]],
                      [-nA[1],  nA[0],     0.]],
                     np.float64)
    curlB = np.array([[    0., -nB[2],  nB[1]],
                      [ nB[2],     0., -nB[0]],
                      [-nB[1],  nB[0],     0.]],
                     np.float64)

    nrmJ = JA0*JA0 + JA1*JA1 + JA2*JA2
    rho = (Ec / (Jc * mu0)) * np.power(nrmJ, (n-1)/2)
    tmp = (n-1) * (Ec / (Jc * mu0)) * np.power(nrmJ, (n-3)/2)
    jac = np.array([[rho + tmp*JA0*JA0,       tmp*JA0*JA1,       tmp*JA0*JA2],
                    [      tmp*JA1*JA0, rho + tmp*JA1*JA1,       tmp*JA1*JA2],
                    [      tmp*JA2*JA0,       tmp*JA2*JA1, rho + tmp*JA2*JA2]],
                   np.float64)

    op = curlA @ jac @ curlB

    row[size] = rowId
    col[size] = colId
    coeff[size] =  alpha * op[2,2] * v
    size += 1

    return 1
@njit
def supLin(row, col, coeff, rowId, colId, size,
           nA, nB, vC, vD, alpha,
           idF, args):
    v   = args[idF][0]
    JA0 = args[idF][1]
    JA1 = args[idF][2]
    JA2 = args[idF][3]

    mu0 = 4 * np.pi * 1e-7
    Ec  = 1e-4
    Jc  =  1e7
    n   =   10

    curlA = np.array([[    0., -nA[2],  nA[1]],
                      [ nA[2],     0., -nA[0]],
                      [-nA[1],  nA[0],     0.]],
                     np.float64)
    curlB = np.array([[    0., -nB[2],  nB[1]],
                      [ nB[2],     0., -nB[0]],
                      [-nB[1],  nB[0],     0.]],
                     np.float64)

    nrmJ = JA0*JA0 + JA1*JA1 + JA2*JA2
    rho = (Ec / (Jc * mu0)) * np.power(nrmJ, (n-1)/2)
    tmp = (n-1) * (Ec / (Jc * mu0)) * np.power(nrmJ, (n-3)/2)
    jac = np.array([[rho + tmp*JA0*JA0,       tmp*JA0*JA1,       tmp*JA0*JA2],
                    [      tmp*JA1*JA0, rho + tmp*JA1*JA1,       tmp*JA1*JA2],
                    [      tmp*JA2*JA0,       tmp*JA2*JA1, rho + tmp*JA2*JA2]],
                   np.float64)

    op = curlA @ jac @ curlB

    for i in range(3):
        for j in range(3):
            row[size] = 3*rowId +i
            col[size] = 3*colId +j
            coeff[size] =  alpha * op[i,j] * v
            size += 1

    return 9
@njit
def supJacZ(row, col, coeff, rowId, colId, size,
              nA, nB, vC, vD, alpha,
              idF, args):
    v   = args[idF][0]
    JA0 = args[idF][1]
    JA1 = args[idF][2]
    JA2 = args[idF][3]

    mu0 = 4 * np.pi * 1e-7
    Ec  = 1e-4
    Jc  =  1e7
    n   =   10

    curlA = np.array([[    0., -nA[2],  nA[1]],
                      [ nA[2],     0., -nA[0]],
                      [-nA[1],  nA[0],     0.]],
                     np.float64)
    curlB = np.array([[    0., -nB[2],  nB[1]],
                      [ nB[2],     0., -nB[0]],
                      [-nB[1],  nB[0],     0.]],
                     np.float64)

    nrmJ = JA0*JA0 + JA1*JA1 + JA2*JA2
    tmp = (n-1) * (Ec / (Jc * mu0)) * np.power(nrmJ, (n-3)/2)
    jac = np.array([[tmp*JA0*JA0, tmp*JA0*JA1, tmp*JA0*JA2],
                    [tmp*JA1*JA0, tmp*JA1*JA1, tmp*JA1*JA2],
                    [tmp*JA2*JA0, tmp*JA2*JA1, tmp*JA2*JA2]],
                   np.float64)

    op = curlA @ jac @ curlB

    row[size] = rowId
    col[size] = colId
    coeff[size] =  alpha * op[2,2] * v
    size += 1

    return 1
@njit
def supJac(row, col, coeff, rowId, colId, size,
              nA, nB, vC, vD, alpha,
              idF, args):
    v   = args[idF][0]
    JA0 = args[idF][1]
    JA1 = args[idF][2]
    JA2 = args[idF][3]

    mu0 = 4 * np.pi * 1e-7
    Ec  = 1e-4
    Jc  =  1e7
    n   =   10

    curlA = np.array([[    0., -nA[2],  nA[1]],
                      [ nA[2],     0., -nA[0]],
                      [-nA[1],  nA[0],     0.]],
                     np.float64)
    curlB = np.array([[    0., -nB[2],  nB[1]],
                      [ nB[2],     0., -nB[0]],
                      [-nB[1],  nB[0],     0.]],
                     np.float64)

    nrmJ = JA0*JA0 + JA1*JA1 + JA2*JA2
    tmp = (n-1) * (Ec / (Jc * mu0)) * np.power(nrmJ, (n-3)/2)
    jac = np.array([[tmp*JA0*JA0, tmp*JA0*JA1, tmp*JA0*JA2],
                    [tmp*JA1*JA0, tmp*JA1*JA1, tmp*JA1*JA2],
                    [tmp*JA2*JA0, tmp*JA2*JA1, tmp*JA2*JA2]],
                   np.float64)

    op = curlA @ jac @ curlB

    for i in range(3):
        for j in range(3):
            row[size] = 3*rowId +i
            col[size] = 3*colId +j
            coeff[size] =  alpha * op[i,j] * v
            size += 1

    return 9

@njit
def getDet(alpha, beta, Jx, Jy, Jz):
    tmp = beta * (Jx*Jx + Jy*Jy + Jz*Jz)
    sol = alpha * alpha + alpha * tmp
    return sol

@njit
def getInv(alpha, beta, Jx, Jy, Jz, matrix):
    tmp = beta * (Jx*Jx + Jy*Jy + Jz*Jz)
    det = alpha * alpha + alpha * tmp

    # det = getDet(alpha, beta, Jx, Jy, Jz)

    matrix[0][0] = (alpha + beta * Jy *Jy + beta * Jz * Jz) / det
    matrix[1][1] = (alpha + beta * Jx *Jx + beta * Jz * Jz) / det
    matrix[2][2] = (alpha + beta * Jx *Jx + beta * Jy * Jy) / det

    matrix[0][1] = - (beta * Jx *Jy) / det
    matrix[0][2] = - (beta * Jx *Jz) / det

    matrix[1][0] = - (beta * Jy *Jx) / det
    matrix[1][2] = - (beta * Jy *Jz) / det

    matrix[2][0] = - (beta * Jz *Jx) / det
    matrix[2][1] = - (beta * Jz *Jy) / det

@njit
def curlCurlLin(row, col, coeff, rowId, colId, size,
                nA, nB, vC, vD, alpha,
                idF, args):

    matrix = args[idF][:]

    curlA = np.array([[    0., -nA[2],  nA[1]],
                      [ nA[2],     0., -nA[0]],
                      [-nA[1],  nA[0],     0.]],
                     np.float64)
    curlB = np.array([[    0., -nB[2],  nB[1]],
                      [ nB[2],     0., -nB[0]],
                      [-nB[1],  nB[0],     0.]],
                     np.float64)

    op = curlA @ matrix @ curlB

    row[size] = rowId
    col[size] = colId
    coeff[size] =  alpha * op[2,2]
    size += 1

    return 1
@njit
def curlCurlLin3(row, col, coeff, rowId, colId, size,
                 nA, nB, vC, vD, alpha,
                 idF, args):

    matrix = args[idF][:]

    curlA = np.array([[    0., -nA[2],  nA[1]],
                      [ nA[2],     0., -nA[0]],
                      [-nA[1],  nA[0],     0.]],
                     np.float64)
    curlB = np.array([[    0., -nB[2],  nB[1]],
                      [ nB[2],     0., -nB[0]],
                      [-nB[1],  nB[0],     0.]],
                     np.float64)

    op = curlA @ matrix @ curlB

    for i in range(3):
        for j in range(3):
            row[size] = 3*rowId +i
            col[size] = 3*colId +j
            coeff[size] =  alpha * op[i,j]
            size += 1

    return 9
